# -*- coding: utf-8 -*-

"""Top-level package for aDMIX."""
__version__ = '1.0.15'

import os
import logging
from utilix import uconfig

def get_logger():
    logger = logging.getLogger("admix")
    ch = logging.StreamHandler()
    ch.setLevel(uconfig.logging_level)
    logger.setLevel(uconfig.logging_level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger

logger = get_logger()

PKGDIR = os.path.dirname(__file__)

DEFAULT_CONFIG = os.path.join(PKGDIR, 'config', 'default.config')

from . import utils
from .downloader import download
from .uploader import upload
from .uploader import preupload
from . import manager
from . import monitor
from . import validator



import os
from argparse import ArgumentParser

import socket

import admix.rucio
from .utils import xe1t_runs_collection
from .rucio import list_rules, get_did_type, get_rses
from .manager import bring_online
from . import logger
from . import clients
try:
    from straxen import __version__
    straxen_version = __version__
except ImportError:
    print("Straxen not installed in current env, so must pass straxen_version manually")
import time
import utilix


class NoRSEForCountry(Exception):
    pass

class RucioDownloadError(Exception):
    pass


def determine_rse(rse_list):
    # TODO put this in config or something?

    preferred_host_rses = {'rcc': ['UC_DALI_USERDISK', 'UC_OSG_USERDISK', 'SDSC_USERDISK', 'SDSC_NSDF_USERDISK'],
                           'sdsc': ['SDSC_USERDISK', 'UC_OSG_USERDISK', 'UC_DALI_USERDISK', 'SDSC_NSDF_USERDISK'],
                           'in2p3': ['CCIN2P3_USERDISK', 'NIKHEF2_USERDISK', 'CNAF_USERDISK'],
                           'nikhef': ['NIKHEF2_USERDISK', 'SURFSARA_USERDISK', 'CNAF_USERDISK'],
                           'surf': ['SURFSARA_USERDISK', 'NIKHEF2_USERDISK', 'CNAF_USERDISK'],
                          }

    preferred_glidein_rses = {'US,CA':  ['UC_OSG_USERDISK', 'SDSC_USERDISK', 'UC_DALI_USERDISK', 'SDSC_NSDF_USERDISK'],
                              'EUROPE,NL,IT,FR,IL': ['NIKHEF2_USERDISK', 'CNAF_USERDISK', 'SURFSARA_USERDISK']
                              }

    hostname = socket.getfqdn()

    # check if we are running on a specific host that's very close to our rses
    for host, pref_rses in preferred_host_rses.items():
        if host in hostname:
            for rse in pref_rses:
                if rse in rse_list:
                    return rse

    # in case we are on an OSG job, check the GLIDEIN_Country
    glidein_country = os.environ.get('GLIDEIN_Country')
    if glidein_country:
        for country_list, pref_rses in preferred_glidein_rses.items():
            country_list = country_list.split(',')
            if glidein_country in country_list:
                for rse in pref_rses:
                    if rse in rse_list:
                        return rse

    # as last ditch effort, default to UC_OSG or SDSC
    for pref_rse in ['UC_OSG_USERDISK', 'SDSC_USERDISK', 'SDSC_NSDF_USERDISK']:
        if pref_rse in rse_list:
            return pref_rse

    # if get here, return None and let rucio figure it out
    return


def download_dids(dids, num_threads=8, **kwargs):
    # build list of did info
    did_list = []
    for did in dids:
        did_dict = dict(did=did,
                        **kwargs
                        )
        did_list.append(did_dict)
    return clients.download_client.download_dids(did_list, num_threads=num_threads)


def download(did, chunks=None, location='.',  tries=3, metadata=True,
             only_metadata=False, num_threads=5, rse=None, stage=None):
    """Function download()

    """
    did_type = get_did_type(did)

    if did_type == 'FILE':
        # make sure we didn't pass certain args
        assert chunks is None, f"You passed the chunks argument, but the DID {did} is a FILE"


    if only_metadata:
        dids = [did + '-metadata.json']
    else:
        if chunks:
            dids = []
            for c in chunks:
                cdid = f"{did}-{c:06d}"
                dids.append(cdid)
            # also download metadata
            if metadata:
                dids.append(did + '-metadata.json')
        else:
            scope = did.split(':')[0]
            dids = [f"{scope}:{f}" for f in admix.rucio.list_files(did)]

    path = did.replace(':', '-')
    # drop the xnt at the beginning
    path = path.replace('xnt_', '')
    if did_type == 'FILE':
        path = '-'.join(path.split('-')[:-1])
    location = os.path.join(location, path)

    os.makedirs(location, exist_ok=True)

    already_exist = []
    already_exist_files = []
    for _did in dids:
        file = os.path.join(location, _did.split(':')[1])
        if os.path.exists(file):
            already_exist.append(_did)
            already_exist_files.append(os.path.abspath(file))

    dids = list(set(dids) - set(already_exist))

    if len(dids) == 0:
        logger.info(f"All files already present at {location}")
        return already_exist_files

    if not rse:
        # determine which rses this did is on
        rses = get_rses(did, state='OK')
        # find closest rse
        rse = determine_rse(rses)


    if stage:
        logger.info(f"Staging {len(dids)} file{'s'*(len(dids)>1)} of {did} from {rse}")        
        bring_online(did,rse)


    if chunks:
        logger.info(f"Downloading {len(dids)} file{'s'*(len(dids)>1)} of {did} from {rse}")
    else:
        logger.info(f"Downloading {did} from {rse}")

    _try = 1
    success = False
    while _try <= tries and not success:
        if _try == tries:
            rse = None
        try:
            result = download_dids(dids, base_dir=location, no_subdir=True, rse=rse, num_threads=num_threads)
            success = True
        except:
            logger.debug(f"Download try #{_try} failed. Sleeping for {3*_try} seconds.")
            time.sleep(3 ** _try)
            _try += 1

    if success:
        logger.debug(f"Download successful to {location}")
    else:
        raise RucioDownloadError(f"Download of {did} failed")

    downloaded_paths = [r['dest_file_paths'][0] for r in result]
    # return list of all files
    paths = sorted(downloaded_paths + already_exist_files)
    return paths

def get_did_1t(number, dtype):
    query = {'number': number}
    cursor = xe1t_runs_collection.find_one(query, {'number': 1, 'name': 1, 'data': 1})
    for d in cursor['data']:
        if dtype == 'raw':
            if d['type'] == 'raw' and d['host'] == 'rucio-catalogue' and d['status'] == 'transferred':
                return d['location']
        else:
            if (d['type'] == 'processed' and d['host'] == 'rucio-catalogue' and d['pax_version'] == dtype and
                d['status'] == 'transferred'):
                return d['location']

    # if get here, there is a problem finding the DID
    raise ValueError(f"No rucio DID found for run {number} with dtype {dtype}")


def download_1t(number, dtype, location='.',  tries=3, num_threads=5, **kwargs):
    did = get_did_1t(number, dtype)

    # if we didn't pass an rse, determine the best one
    rse = kwargs.pop('rse', None)

    if not rse:
        rules = list_rules(did, state='OK')
        rses = [r['rse_expression'] for r in rules]
        # find closest rse
        rse = determine_rse(rses)

    if dtype == 'raw':
        # get run name
        name = xe1t_runs_collection.find_one({'number': number, 'detector': 'tpc'}, {'name': 1})['name']
        location = os.path.join(location, name)

    os.makedirs(location, exist_ok=True)

    print(f"Downloading {did} from {rse}")

    _try = 1
    success = False

    while _try <= tries and not success:
        if _try == tries:
            rse = None
        result = download_dids(did, base_dir=location, no_subdir=True, rse=rse, num_threads=num_threads)
        if isinstance(result, int):
            print(f"Download try #{_try} failed.")
            time.sleep(5**_try)
            _try += 1

        else:
            success = True

    if success:
        print(f"Download successful to {location}")
    else:
        raise RucioDownloadError("Download of {did} failed")



"""
Utility functions/objects used by admix generally
"""
from utilix import xent_collection, xe1t_collection, DB

from . import logger

try:
    db = DB()
except:
    print("DB initialization failed")
    logger.warning(f"Initializing utilix DB failed. You cannot do database operations")
    db = None


xent_runs_collection = xent_collection()
xent_context_collection = xent_collection('contexts')
xe1t_runs_collection = xe1t_collection()


RAW_DTYPES = ['raw_records',
              'raw_records_he',
              'raw_records_mv',
              'raw_records_nv',
              'raw_records_aqmon',
              'raw_records_aux_mv',
              'raw_records_aqmon_nv'
              ]

RAWish_DTYPES = RAW_DTYPES + ['records',
                              'records_he',
                              'records_nv',
                              'raw_records_coin_nv',
                              'lone_raw_records_nv',
                              'lone_raw_record_statistics_nv',
                              'records_mv'
                              ]


def make_did(run_number, dtype, hash):
    ### HUGE WARNING: DO NOT CHANGE THIS FUNCTION!!!!
    ## IT WILL BREAK LITERALLY EVERYTHING
    scope = 'xnt_%06d' % run_number
    dataset = "%s-%s" % (dtype, hash)
    return "%s:%s" % (scope, dataset)
    ### HUGE WARNING: DO NOT CHANGE THIS FUNCTION!!!!


def parse_did(did):
    """Takes a did and returns the run number, dtype, hash"""
    scope, name = did.split(':')
    number = int(scope.split('_')[1])
    dtype, h = name.split('-')
    return number, dtype, h


def make_highlevel_container_did(run_number, straxen_version):
    scope = 'xnt_%06d' % run_number
    if not straxen_version.startswith('v'):
        straxen_version = 'v' + straxen_version
    container_name = 'highlevel_' + straxen_version.replace('.', '-')
    return f"{scope}:{container_name}"


def parse_dirname(dirname):
    number, dtype, lineage_hash = dirname.split('-')
    number = int(number)
    return number, dtype, lineage_hash





from admix.manager import synchronize
from .daemon import AdmixDaemon


class SyncDaemon(AdmixDaemon):
    """Syncs the RunDB with Rucio"""

    query = {'data': {'$elemMatch': {'status': 'transferring'}},
             'number': {'$gt': 10000}
             }

    desc = "Syncing rundb with rucio"

    def __init__(self, db_query=None, dtype=None):
        super().__init__(db_query)
        self.dtype = dtype

    def do_task(self, rundoc):
        synchronize(rundoc['number'], self.dtype)




"""
Contains rucio commands with XENON-specific wrappers
"""
from functools import wraps
import re
from tqdm import tqdm
import rucio.common.exception

import admix.utils
from . import logger
from .utils import parse_did, db, RAW_DTYPES
from . import clients


needs_client = clients.needs_client


def requires_production(func):
    """Decorator used for functions that require the production account credentials"""
    def wrapped(*args, **kwargs):
        if clients.rucio_client.account != 'production':
            raise RucioPermissionError(f"You must be the production user to call {func.__name__}")
        return func(*args, **kwargs)
    return wrapped


def build_data_dict(did, rse, status):
    number, dtype, h = parse_did(did)
    files = list_files(did, verbose=True)
    size = sum([f['bytes'] for f in files]) / 1e6
    data = dict(did=did,
                type=dtype,
                location=rse,
                status=status,
                host='rucio-catalogue',
                meta=dict(lineage_hash=h,
                          size_mb=size,
                          file_count=len(files)
                          ),
                protocol='rucio',
                )
    return data


def update_db(mode):
    """Mode can be 'add', 'move', 'delete'"""
    def decorator(func):
        @wraps(func)
        def wrapped(*args,  **kwargs):
            # did and rse are the first 2 args
            did, rse = args[0], args[1]
            update_db = kwargs.get('update_db', False)
            if not update_db or db is None:
                return func(*args, **kwargs)

            func(did, rse, **kwargs)
            number, dtype, h = parse_did(did)
            if mode == 'add':
                data = build_data_dict(did, rse, status='transferring')
                db.update_data(number, data)

            elif mode == 'delete':
                data = db.get_data(number, type=dtype, did=did, location=rse)
                if len(data) == 0:
                    pass
                else:
                    for d in data:
                        db.delete_data(number, d)

            elif mode == 'move':
                from_rse = kwargs.pop('from_rse')
                if not from_rse:
                    raise ValueError(f"'from_rse' must be passed when calling {func.__name__}")
                from_data = db.get_data(number, type=dtype, did=did, location=from_rse)
                if len(from_data) == 0:
                    to_data = build_data_dict(did, rse, 'transferring')
                else:
                    to_data = from_data[0].copy()
                    to_data['location'] = rse
                    to_data['status'] = 'transferring'
                db.update_data(number, to_data)
                db.delete_data(number, from_data)
        return wrapped
    return decorator


@needs_client
def get_did(did):
    scope, name = did.split(':')
    try:
        return clients.rucio_client.get_did(scope, name)
    except rucio.common.exception.DataIdentifierNotFound:
        pass


@needs_client
def get_did_type(did):
    scope, name = did.split(':')
    return clients.rucio_client.get_did(scope, name)['type']


@needs_client
def list_rules(did, **filters):
    scope, name = did.split(':')
    # check if did is a file or a dataset
    if get_did_type(did) == 'FILE':
        rules = clients.rucio_client.list_associated_rules_for_file(scope, name)

    else:
        rules = clients.rucio_client.list_did_rules(scope, name)
    # get rules that pass some filter(s)
    ret = []
    for rule in rules:
        selected = True
        for key, val in filters.items():
            if rule[key] != val:
                selected = False
        if selected:
            ret.append(rule)
    return ret

@needs_client
def get_rses(did, **filters):
    rules = list_rules(did, **filters)
    rses = [r['rse_expression'] for r in rules]
    return rses

@needs_client
def get_rule(did, rse):
    rules = list_rules(did, rse_expression=rse)
    if len(rules) == 0:
        logger.debug(f"No rule found for {did} at {rse}!")
        return
    return rules[0]


@needs_client
@requires_production
@update_db('add')
def add_rule(did, rse, copies=1, update_db=False, quiet=False, **kwargs):
    scope, name = did.split(':')
    did_dict = dict(scope=scope, name=name)
    try:
        clients.rucio_client.add_replication_rule([did_dict], copies, rse_expression=rse, **kwargs)
    except rucio.common.exception.DuplicateRule:
        if not quiet:
            print(f"Rule already exists for {did} at {rse}")
        return
    if not quiet:
        print(f"Replication rule added for {did} at {rse}")


@needs_client
@requires_production
@update_db('delete')
def delete_rule(did, rse, purge_replicas=True, _careful=True, _required_copies=1, update_db=False,
                quiet=False):
    number, dtype, hsh = parse_did(did)
    if dtype in RAW_DTYPES and _required_copies < 1:
        raise DataPolicyError("You cant remove raw_records data. Shame on you. ")
    rule = get_rule(did, rse)
    if not rule:
        raise RuleNotFoundError(f"No rule found for {did} at {rse}!")
    # make sure there is at least one other copy
    if _careful:
        rules = list_rules(did, state='OK')
        other_rules = [r for r in rules if r['rse_expression'] != rse and
                       r['expires_at'] is None
                       ]
        if len(other_rules) < _required_copies:
            raise DataPolicyError(f"We require at least {_required_copies} long-term copies "
                                  f"and deleting one for {did} would result in {len(other_rules)}."
                                  )
    clients.rucio_client.delete_replication_rule(rule['id'], purge_replicas=purge_replicas)
    if not quiet:
        print(f"Replication rule for {did} at {rse} removed.")


@needs_client
@requires_production
def erase(did, now=False, update_db=False):
    scope, name = did.split(':')
    if get_did_type(did) == "FILE":
        dtype = name.split('-')[0]
    else:
        number, dtype, hsh = parse_did(did)
    if dtype in admix.utils.RAW_DTYPES:
        print(f"You cannot erase {dtype} data. Shame on you")
        return
    # delete DID in 10 seconds if pass now=True, else copy what rucio does and set it to 24 hours.
    # see https://github.com/rucio/rucio/blob/master/bin/rucio#L883
    value = 10 if now else 86400
    if update_db and get_did_type(did) == "DATASET":
        number, dtype, h = parse_did(did)
        data = db.get_data(number, did=did)
        for d in data:
            db.delete_data(number, d)
    try:
        clients.rucio_client.set_metadata(scope, name, key='lifetime', value=value)
    except rucio.common.exception.DataIdentifierNotFound:
        print(f"{did} does not exist")


@requires_production
@update_db('add')
def add_conditional_rule(did, rse, from_rse=None, update_db=False, **kwargs):
    """Convenience wrapper around add_rule"""
    add_rule(did, rse, source_replica_expression=from_rse, **kwargs)


@requires_production
@update_db('move')
def move_rule(did, rse, from_rse, update_db=False):
    pass


@needs_client
def add_scope(account, scope):
    return clients.rucio_client.add_scope(account, scope)


@needs_client
@requires_production
def add_production_scope(scope):
    return add_scope('production', scope)


@needs_client
@requires_production
def add_container(scope, name, **kwargs):
    return clients.rucio_client.add_container(scope, name, **kwargs)


@needs_client
def list_datasets(scope):
    datasets = [d for d in clients.rucio_client.list_dids(scope, {'type': 'dataset'}, type='dataset')]
    return datasets


@needs_client
def list_containers(scope):
    containers = [d for d in clients.rucio_client.list_dids(scope, {'type': 'container'}, type='container')]
    return containers


@needs_client
def list_scopes(regex_pattern='.*'):
    pattern = re.compile(regex_pattern)
    _scopes = clients.rucio_client.list_scopes()
    scopes = [s for s in _scopes if pattern.match(s)]
    return scopes


@needs_client
def list_content(did, full_output=False):
    # if full_output is False (default), just return a list of content names in the DID
    # otherwise, return a list of dicts with everythign rucio sends back
    scope, name = did.split(':')
    if full_output:
        content = [d for d in clients.rucio_client.list_content(scope, name)]
    else:
        content = [f"{d['scope']}:{d['name']}" for d in clients.rucio_client.list_content(scope, name)]
    return content


@needs_client
def list_files(did, verbose=False):
    scope, name = did.split(':')
    if verbose:
        files = [f for f in clients.rucio_client.list_files(scope, name)]
    else:
        files = [f['name'] for f in clients.rucio_client.list_files(scope, name)]
    return files


@needs_client
def attach(main_did, attachments, rse=None):
    # attach a list of attachments to the main_did, either a dataset or container
    # the attachments are a list of DIDs
    attachment_dicts = []
    for did in attachments:
        _scope, _name = did.split(':')
        attachment_dicts.append(dict(scope=_scope, name=_name))
    main_scope, main_name = main_did.split(':')
    return clients.rucio_client.attach_dids(main_scope, main_name, attachment_dicts)


@needs_client
def get_size_mb(did):
    # returns size of did (or list of dids) in GB
    if not isinstance(did, str):
        # then assume it's iterable
        total_size = sum([get_size_mb(d) for d in tqdm(did, desc='Getting total size in rucio')])
    else:
        if not isinstance(did, str):
            raise ValueError(f"did must be a string (or an iterable of strings). You passed a {type(did)}")
        scope, name = did.split(':')
        total_size = 0
        for f in clients.rucio_client.list_files(scope, name):
            total_size += int(f['bytes'])/1e6
    return total_size


@needs_client
def list_file_replicas(did, rse=None, **kwargs):
    if 'rse_expression' in kwargs:
        rse_expression = kwargs.pop('rse_expression')
        if kwargs['rse'] != rse:
            raise ValueError(f"You passed rse={rse} and rse_expression={rse_expression}. Pick one.")
    scope, name = did.split(':')
    did_dict = [dict(scope=scope, name=name)]
    replicas = clients.replica_client.list_replicas(did_dict, rse_expression=rse, **kwargs)
    ret = []
    for r in replicas:
        d = dict(name=r['name'], rses=r['rses'])
        ret.append(d)
    return ret


@needs_client
def list_rses():
    return [rse for rse in clients.rse_client.list_rses()]

@needs_client
def get_rse_usage(rse):
    usage = next(clients.rse_client.get_rse_usage(rse))['used']
    return usage

@needs_client
def get_account_usage(account='production', rse=None):
    keep_fields = ['files', 'bytes', 'bytes_limit', 'bytes_remaining']
    if rse is None:
        ret = []
        for rse in list_rses():
            ret.append(get_account_usage(account, rse['rse']))
        return ret
    else:
        usage = clients.account_client.get_local_account_usage(account, rse=rse)
        try:
            usage = next(usage)
            for field in list(usage.keys()):
                if field not in keep_fields:
                    usage.pop(field)
        except StopIteration:
            usage = {field: 0 for field in keep_fields}
        return usage

@needs_client
def get_rse_prefix(rse):
    rse_info = clients.rse_client.get_rse(rse)
    prefix = rse_info['protocols'][0]['prefix']
    return prefix


@needs_client
def get_rse_datasets(rse):
    datasets = clients.replica_client.list_datasets_per_rse(rse)
    ret = []
    for d in tqdm(datasets, desc=f'Finding all datasets at {rse}'):
        ret.append(f"{d['scope']}:{d['name']}")
    return ret


class RucioPermissionError(Exception):
    pass


class RuleNotFoundError(Exception):
    pass


class DataPolicyError(Exception):
    pass




"""
Common data management functions
"""
import os
import shutil
import tempfile

from packaging import version
from tqdm import tqdm
from rucio.common.exception import DataIdentifierNotFound

import admix.rucio
from . import clients
from . import rucio, logger
from . import utils
from .utils import db

import gfal2
import time
from datetime import timezone, datetime, timedelta

TAPE_RSES = ['SURFSARA_USERDISK','CNAF_TAPE3_USERDISK','CNAF_TAPE2_USERDISK','CNAF_TAPE_USERDISK','CCIN2P3_USERDISK']

def has_metadata(did):
    scope, dset = did.split(':')
    files = rucio.list_files(did)
    metadata_file = f"{dset}-metadata.json"
    return metadata_file in files


def synchronize(run_number, dtype=None):
    # use dtype a lot below so set the arg to _dtype
    _dtype = dtype
    del dtype
    if _dtype is not None:
        db_data = db.get_data(run_number, host='rucio-catalogue', type=_dtype)
    else:
        db_data = db.get_data(run_number, host='rucio-catalogue')

    try:
        db_datasets = list(set([d['did'].split(':')[1] for d in db_data]))
    except KeyError:
        db_datasets = []

    scope = f'xnt_{run_number:06d}'
    rucio_datasets = rucio.list_datasets(scope)
    if _dtype is not None:
        rucio_datasets = [ds for ds in rucio_datasets if ds.split('-')[0] == _dtype]

    # find datasets that are in rucio but not in DB
    missing_db = set(rucio_datasets) - set(db_datasets)

    # add new datasets to runDB that are missing
    for dset in missing_db:
        did = f"{scope}:{dset}"

        dtype, strax_hash = dset.split('-')

        # get the locations of this dataset
        rules = rucio.list_rules(did)

        for rule in rules:
            if rule['state'] == 'OK':
                status = 'transferred'
            elif rule['state'] == 'REPLICATING':
                status = 'transferring'
            elif rule['state'] == 'STUCK':
                status = 'error'
            elif rule['state'] == 'SUSPENDED':
                status = 'error'
            else:
                raise ValueError(f"rule {rule['id']} is in an unknown state: {rule['state']}")

            location = rule['rse_expression']

            # if state is OK, do one last sanity check that the metadata is there
            # this is edge case I found during outsource once
            if status == 'transferred':
                if not has_metadata(did):
                    status = 'transferring'

            new_datum = dict(host='rucio-catalogue',
                             did=did,
                             type=dtype,
                             status=status,
                             location=location,
                             protocol='rucio'
                             )
            logger.debug(f"Adding {new_datum['did']} at {new_datum['location']}")
            db.update_data(run_number, new_datum)

    # now modify/remove runDB entries according to rucio
    for dset in db_datasets:
        did = f"{scope}:{dset}"

        copies = [d for d in db_data if d['did'] == did]

        base_dict = copies[0].copy()
        base_dict['location'] = None
        base_dict['status'] = None

        db_rses = [d['location'] for d in copies]
        try:
            rucio_rules = rucio.list_rules(did)
            rucio_rses = [r['rse_expression'] for r in rucio_rules]
        except DataIdentifierNotFound:
            rucio_rses = []

        # check for dupicate entries in the runDB
        duplicated_rses = []
        if len(db_rses) != len(set(db_rses)):
            duplicated_rses = list(set([rse for rse in db_rses if db_rses.count(rse) > 1]))

        rses_rm_from_db = set(db_rses) - set(rucio_rses)
        rses_add_to_db = set(rucio_rses) - set(db_rses)
        rses_common = set(db_rses) & set(rucio_rses)

        for rse in rses_rm_from_db:
            pull_data = [d for d in copies if d['location'] == rse][0]
            logger.debug(f"Removing db {pull_data['did']} at {pull_data['location']}")
            db.delete_data(run_number, pull_data)

        for rse in rses_add_to_db:
            rule = [r for r in rucio_rules if r['rse_expression'] == rse][0]
            if rule['state'] == 'OK':
                status = 'transferred'
            elif rule['state'] == 'REPLICATING':
                status = 'transferring'
            elif rule['state'] == 'STUCK':
                status = 'stuck'
            elif rule['state'] == 'SUSPENDED':
                status = 'error'
            else:
                raise ValueError(f"rule {rule['id']} is in an unknown state: {rule['state']}")

            # if state is OK, do one last sanity check that the metadata is there
            # this is edge case I found during outsource once
            if status == 'transferred':
                if not has_metadata(did):
                    status = 'transferring'

            new_datum = base_dict.copy()
            new_datum['location'] = rse
            new_datum['status'] = status
            logger.debug(f"Adding {new_datum['did']} at {new_datum['location']}")
            db.update_data(run_number, new_datum)

        for rse in rses_common:
            db_datum = [d for d in copies if d['location'] == rse][0]

            rule = [r for r in rucio_rules if r['rse_expression'] == rse][0]

            if rule['state'] == 'OK':
                status = 'transferred'
            elif rule['state'] == 'REPLICATING':
                status = 'transferring'
            elif rule['state'] == 'STUCK':
                status = 'stuck'
            elif rule['state'] == 'SUSPENDED':
                status = 'error'
            else:
                raise ValueError(f"rule {rule['id']} is in an unknown state: {rule['state']}")

            # if state is OK, do one last sanity check that the metadata is there
            # this is edge case I found during outsource once
            if status == 'transferred':
                if not has_metadata(did):
                    logger.debug(f"Warning! no metadata found for {did}")
                    status = 'transferring'

            if db_datum['status'] != status:
                updatum = db_datum.copy()
                updatum['status'] = status
                logger.debug(f"Updating {updatum['did']} at {updatum['location']}")
                db.update_data(run_number, updatum)

        if len(duplicated_rses):
            for rse in duplicated_rses:
                logger.debug(f"We have duplicate entries for {did} at {rse}.")
                # let's just remove all of them and then add back in
                ddocs = [ddoc for ddoc in copies if ddoc['location'] == rse]

                for ddoc in ddocs:
                    db.delete_data(run_number, ddoc)

            synchronize(run_number)


def add_rucio_protocol(run_number):
    """Straxen requires for some reason a field protocol=rucio in the data dict"""
    data = db.get_data(run_number, host='rucio-catalogue')
    for d in data:
        if d.get('protocol', 'no_protocol') != 'rucio':
            updatum = d.copy()
            updatum['protocol'] = 'rucio'
            logger.debug(f"Updating {d['did']} at {d['location']}")
            db.update_data(run_number, updatum)


def get_outdated_strax_info(not_outdated_version,
                            return_current_hashes=False
                            ):
    coll = utils.xent_context_collection

    # get versions of all straxen versions before some given 'good' version
    thresh_version = version.parse(not_outdated_version)

    cursor = list(coll.find({}, {'straxen_version': 1}))
    versions = set([version.parse(d['straxen_version']) for d in cursor])
    outdated_versions = set(sorted([v for v in versions if v < thresh_version], reverse=True))
    save_versions = versions - outdated_versions
    # if no save_versions, then save the data of the latest version
    if not len(save_versions):
        v = sorted(list(versions))[-1]
        save_versions = [v]
        outdated_versions.remove(v)

    # get the hashes of all datatypes we want to save
    save_hashes = dict()
    for v in save_versions:
        cursor = coll.find({'straxen_version': v.public})
        for doc in cursor:
            hashes = doc['hashes']
            for dtype, h in hashes.items():
                if dtype in save_hashes:
                    if h not in save_hashes[dtype]:
                        save_hashes[dtype].append(h)
                else:
                    save_hashes[dtype] = [h]

    delete_hashes = dict()
    for v in outdated_versions:
        cursor = coll.find({'straxen_version': v.public})
        for doc in cursor:
            hashes = doc['hashes']
            for dtype, h in hashes.items():
                if dtype in save_hashes and h in save_hashes[dtype]:
                    pass
                else:
                    if dtype in delete_hashes:
                        if h not in delete_hashes[dtype]:
                            delete_hashes[dtype].append(h)
                    else:
                        delete_hashes[dtype] = [h]

    # just to double check
    for dtype, h in delete_hashes.items():
        if dtype in save_hashes:
            assert h not in save_hashes[dtype]
    if return_current_hashes:
        return delete_hashes, save_hashes
    return delete_hashes


def find_outdated_data(max_straxen_version, specific_dtype=None):
    def get_dids(ddoc, dtype, hash_list):
        ret = []
        for d in ddoc:
            if d['type'] == dtype and d.get('did'):
                ddoc_hash = d['did'].split(':')[1].split('-')[1]
                if ddoc_hash in hash_list:
                    if d['did'] not in ret:
                        ret.append(d['did'])
        if not len(ret):
            raise RuntimeError(f"No dids found for {dtype}")
        return ret

    outdated_info = get_outdated_strax_info(max_straxen_version)
    if specific_dtype:
        if isinstance(specific_dtype, str):
            outdated_info = {specific_dtype: outdated_info[specific_dtype]}
        else:
            outdated_info = {key: val for key, val in outdated_info.items() if key in specific_dtype}

    outdated_dids = dict()
    for dtype, hsh_list in tqdm(outdated_info.items(), desc='Finding data we can delete'):
        query = {'$or': [{'data': {'$elemMatch': {'type': dtype,
                                                  'did': {'$regex': h}}}}
                         for h in hsh_list]
                 }
        ndocs = utils.xent_runs_collection.count_documents(query)
        if ndocs > 0:
            cursor = utils.xent_runs_collection.find(query, {'number': 1, 'data': 1})
            if dtype not in outdated_dids:
                outdated_dids[dtype] = list()
            for run in cursor:
                dids = get_dids(run['data'], dtype, hsh_list)
                outdated_dids[dtype].extend(dids)
    return outdated_dids


def copy_high_level_data(runlist, move_dont_copy=False):
    """Copy/move all high level data for a given runlist"""
    pass


@rucio.requires_production
def containerize(run_number, straxen_version, context='xenonnt_online',
                 rse=None):
    # get the dtype-hash pairs for this context and this straxen version
    hashes = db.get_context(context, straxen_version)['hashes']

    # get the data entries
    data = db.get_data(run_number, host='rucio-catalogue')

    # loop over data entries, find high-level  ones, and add to a list
    # this list will be used to create the new container
    high_level_dids = []
    for d in data:
        if d['type'] not in utils.RAWish_DTYPES:
            if hashes.get(d['type']) in d.get('did', ''):
                if d['did'] not in high_level_dids:
                    high_level_dids.append(d['did'])

    # if there are not any high-level DIDs, nothing to do
    if not len(high_level_dids):
        print(f"No high-level DIDs found for run {run_number}, straxen {straxen_version}, context {context}")
        return

    container_did = utils.make_highlevel_container_did(run_number, straxen_version)

    # if container does not exist, make it
    scope, container_name = container_did.split(':')
    existing_containers = rucio.list_containers(scope)
    if container_name not in existing_containers:
        print(f"Creating container {container_did}")
        rucio.add_container(scope, container_name)

    # now loop over highlevel datasets, and add non-existing ones to the container
    existing_content = rucio.list_content(container_did)

    to_attach = []
    for dset_did in high_level_dids:
        if dset_did not in existing_content:
            to_attach.append(dset_did)

    if len(to_attach):
        print("Attaching", to_attach)
        rucio.attach(container_did, to_attach, rse=rse)
    else:
        print(f"Nothing to attach for run {run_number}, straxen {straxen_version}, context {context}")


def clean_local_dir(path, before_straxen_version, ensure_rucio=False, dry_run=True):

    """Checks a local directory for any outdated straxen version and deletes it.

    :param: path: Local data path to check for outdated data.
    :param: before_straxen_version: Delete any data OLDER than this version. This version will not be deleted.
    :param: ensure_rucio: check if data is uploaded to rucio before deleting.
    :param: dry_run: When True, do not delete anything -- just compile a list of the datasets that can be deleted and write to stdout.
    """

    # get context names from the runDB
    cursor = utils.xent_context_collection.find({}, {'name': 1})
    contexts = set([c['name'] for c in cursor if c.get('name')])

    save_hashes = {}
    for context in contexts:
        _delete_hashes, _save_hashes = get_outdated_strax_info(before_straxen_version,
                                                               context=context,
                                                               return_current_hashes=True)
        for dtype, hash_list in _save_hashes.items():
            if dtype in save_hashes:
                save_hashes[dtype] = save_hashes[dtype] | set(hash_list)

            else:
                save_hashes[dtype] = set(hash_list)


    # only used for when dry_run = True
    to_delete = []
    desc = f"Finding outdated data in {path}" if dry_run else f"Deleting outdated data in {path}"
    for dataset in tqdm(os.listdir(path), desc=desc):
        number, dtype, lineage_hash = utils.parse_dirname(dataset)
        if not save_hashes.get(dtype) or lineage_hash in save_hashes[dtype]:
            # if dtype not in save_hashes, assume it is from another context (so dont delete)
            # if is in save_hashes but the specific hash is in a version to save, dont delete
            continue
        # if we get here, then this dtype-hash pair can be deleted in principle
        if ensure_rucio:
            # check that the data is in rucio first
            # do direct rucio query lest we rely on runDB-rucio synchronization
            did = utils.make_did(number, dtype, lineage_hash)
            try:
                rules = admix.rucio.get_rses(did, state="OK")
                if not len(rules):
                    continue
            except DataIdentifierNotFound:
                continue

        if dry_run:
            # dont delete now, just add to a running list we will output later
            to_delete.append(dataset)
        else:
            shutil.rmtree(os.path.join(path, dataset))

    if dry_run:
        _, tmpfile = tempfile.mkstemp(suffix='.txt', dir="./")
        with open(tmpfile, 'w') as f:
            for dataset in to_delete:
                f.write(f"{os.path.join(path, dataset)}\n")
        print(f"Files that we can delete written to {tmpfile}")


def bring_online(did,rse,timeout=3600):

    """Stages data belonging to a given did from a given rse. 

    :param did: DID to stage
    :param rse: from which RSE data have to be stages
    :param timeout: optionally you can change the maximum time after which the function gives up waiting for full data staging
    :return: True if successfull or if staging is not needed, False in case of error or timeout
    """

    print("Bringing online {0} from {1}".format(did,rse))
    if timeout>600:
        print("** Warning! ** This function could take some time (within a timeout equal to {0} seconds). Consider running your program in a screen session".format(timeout))

    if rse not in TAPE_RSES:
        print("{0} is not a tape-based RSE. Staging is not required".format(rse))
        return(True)

    scope = did.split(':')[0]
    dataset = did.split(':')[1]

    file_replicas = clients.replica_client.list_replicas([{'scope':scope,'name': dataset}],rse_expression=rse)
#    file_replicas = rucio.list_file_replicas(did,rse=rse)
    files = [list(replica['pfns'].keys())[0] for replica in file_replicas]
#    files = rucio.list_file_replicas(did,rse=rse)

    print("Bringing online {0} files".format(len(files)))

    if rse=="SURFSARA_USERDISK":
        for i, file in enumerate(files):
            files[i] = files[i].replace("gsiftp","srm")
            files[i] = files[i].replace("gridftp","srm")
            files[i] = files[i].replace("2811","8443")
    if rse=="CCIN2P3_USERDISK":
        for i, file in enumerate(files):
            files[i] = files[i].replace("gsiftp","srm")
            files[i] = files[i].replace("ccdcacli392.in2p3.fr","ccsrm02.in2p3.fr")
            files[i] = files[i].replace("2811","8443")

    #print(files)                                                                                                                                                                                                                       
    ctx = gfal2.creat_context()

    try:
        # bring_online(surls, pintime, timeout, async)                                                                                                                                                                                  
        # Parameters:                                                                                                                                                                                                                   
        #   surls is the given [srmlist] argument                                                                                                                                                                                       
        #   pintime in seconds (how long should the file stay PINNED), e.g. value 1209600 will pin files for two weeks                                                                                                                  
        #   timeout of request in seconds, e.g. value 604800 will timeout the requests after a week                                                                                                                                     
        #   async is asynchronous request (does not block if != 0)                                                                                                                                                                      
        pintime = 3600*48
        timeout_request = 3600
        (status, token) = ctx.bring_online(files, pintime, timeout_request, True)
        if token:
            print(("Got token %s" % token))
        else:
            print("No token was returned. Are all files online?")
    except gfal2.GError as e:
        print("Could not bring the files online:")
        print(("\t", e.message))
        print(("\t Code", e.code))

    print("Waiting until they are all online... (this might take time)")
    start_time = datetime.now().replace(tzinfo=timezone.utc)
    while True:
        errors = ctx.bring_online_poll(files, token)
        ncompleted = 0
        for surl, error in zip(files, errors):
            if not error:
                ncompleted += 1
        print("So far {0} files have been staged".format(ncompleted))
        if ncompleted == len(files):
            print("Staging of {0} files successfully completed".format(ncompleted))
            return(True)
        now_time = datetime.now().replace(tzinfo=timezone.utc)
        delta_time = now_time - start_time
        if delta_time > timedelta(seconds = timeout):
            print("Timeout reached. Stopping waiting for full data staging")
            print("** Warning ** Staging is not yet completed. If you plan to access data now, performances might be degraded")
            return(False)
        time.sleep(60)



from ..utils import RAW_DTYPES, make_did
from ..uploader import upload
from .daemon import AdmixDaemon


class UploadDaemon(AdmixDaemon):
    # TODO fix this, just a guess
    query = {'status': 'eb_ready_to_upload', 'bootstrax.state': 'done'}

    projection = {'number': 1, 'data': 1, 'bootstrax': 1, 'tags': 1}
    rse = 'LNGS_USERDISK'

    def do_task(self, rundoc):
        # get the datatypes that are on EB machines
        for data_doc in rundoc['data']:
            if 'eb' in data_doc['host']:
                # get did for unique identifier
                did = make_did(rundoc['run_number'], data_doc['type'], data_doc['meta']['lineage_hash'])

                # check if this did is in rucio
                already_uploaded = False
                for other_data_doc in rundoc['data']:
                    if other_data_doc['host'] == 'rucio-catalogue' and other_data_doc['did']==did:
                        already_uploaded = True

                # upload if needed
                if not already_uploaded:
                    self.do_upload(rundoc['number'], data_doc)


    def do_upload(self, run_number, data_doc):
        """Given a data_doc of where some data is on EBs, upload to rucio"""
        # TODO
        # get path of data to upload
        path = data_doc['location']
        rse = self.rse
        did = make_did(run_number, data_doc['type'], data_doc['meta']['lineage_hash'])

        # THIS DOESNT WORK YET, JUST DRAFTING SOME STUFF
        # FIXME -- need to modify upload function to automatically update runDB
        upload(path, rse, did=did, update_db=True)



"""
Validate that the data can be loaded successfully and move the data to the production folder.

Copied shamelessly from reprox: https://github.com/XENONnT/reprox/blob/master/reprox/validate_run.py

But removed strax + reprox dependency and made a function instead of a class
"""

import typing as ty
import json
from warnings import warn
import os
import shutil
import glob
from collections import defaultdict


class StraxDataValidation:
    """
    Check that a directory (corresponding to a single datatype is
    """

    def __init__(self, path: str):
        self.path = path

    def find_error(self) -> str:
        """Run several checks on a path to see if the processing was done correctly"""
        if self._is_temp():
            return 'is_temp_folder'

        md = self._open_metadata()
        if not md:
            return 'has_no_metadata'

        if self._did_fail(md):
            return 'has_exception'

        if self._misses_chunks(md):
            return 'misses_chunks'

        key = os.path.split(self.path)[-1]
        split_key = key.split('-')

        if self._wrong_format(split_key):
            return 'is_wrong_format'

        return False

    def _is_temp(self):
        if '_temp' in self.path:
            warn(f'{self.path} is not finished', UserWarning)
            return True

    def _did_fail(self, md):
        if 'exception' in md:
            warn(f'{self.path} has an exception', UserWarning)
            return True

    def _misses_chunks(self, md):
        n_files = os.listdir(self.path)
        n_chunks = len(n_files) - 1  # metadata
        chunks = [c.get('n') > 0 for c in md['chunks']]
        if n_chunks != sum(chunks):
            warn(f'{self.path} misses chunks?!', UserWarning)
            return True

    @staticmethod
    def _wrong_format(split_key):
        if len(split_key) != 3:
            warn(f'{split_key} is not correct format?!', UserWarning)
            return True
        return False

    def _open_metadata(self) -> dict:
        files = glob.glob(os.path.join(self.path, '*'))
        for f in files:
            if 'metadata' in f:
                md_path = f
                break
        else:
            return {}

        with open(md_path, mode='r') as f:
            return json.loads(f.read())



# -*- coding: utf-8 -*-
from __future__ import with_statement
import os
import json
import datetime
import time
import numpy as np
from .. import DEFAULT_CONFIG

#Create global variables:
global_dictionary = {}

def make_global(dict_key, dict_value):
    """Function: make_global
    Provides a global dictionary across aDMIX
    :param dict_key: The name of the variable
    :param dict_value: The value of the variable
    """
    global global_dictionary

    if dict_key not in global_dictionary:
        global_dictionary[dict_key]=dict_value


def get_hostconfig(key=None):
    with open(global_dictionary.get('admix_config', DEFAULT_CONFIG), 'r') as data_file:
        data = json.load(data_file)
        if key in data:
            return data[key]
        else:
            return data


def get_hostname():
    return os.environ.get('HOSTNAME')


def run_number_converter_full(run_number=None):
    #convert the run number input from terminal by two
    #operators:
    #  - The ',' separates the individual run numbers
    #  - The '-' separates the individual run numbers by
    #            to create a sequence between run numbers
    nb_array = []
    if run_number != None:
        #splits by ',':
        #get all run numbers which are given by
        #commandline before possible sequence operator
        #(split without, makes no trouble)
        rn = run_number.split(",")

        #split individual array entries by '-'
        #get a continous run number sequence each

        for i_rn in rn:
            if i_rn.find('-') >= 0:
                j_rn = i_rn.split("-")
                j_rnA = int(j_rn[0])
                j_rnB = int(j_rn[1])
                j_rnArray = np.arange(j_rnA, j_rnB+1)
                nb_array.extend( j_rnArray )
            else:
                nb_array.append(int(i_rn))
    else:
        nb_array = None

    return nb_array


def eval_run_numbers(run_numbers=None, run_number_min=None, run_number_max=None):
    """Function: eval_run_numbers
    Deals with several ways to hand over the timestamp/run names in aDMIX
    1) <date>_<time> (Single)
    2) <date>_<time>,<date>_<time> (Multiple)
    :param timestamp: A string of timestamps what follows single, multiple criterion
    :return list: A list of timestamps
    """
    #test if run_numbers follows a certain structure:
    eval_nb_min = None
    eval_nb_max = None
    if run_numbers != None and run_number_min != None and run_number_max != None:

        if run_numbers.find("-") > 0:
            rn_beg, rn_end = run_numbers.split("-")

            if rn_beg.isdigit():
                eval_nb_min = rn_beg
            elif rn_beg.isdigit() == False and rn_beg == 'MIN':
                eval_nb_min = run_number_min
            else:
                #todo raise exception
                print("Check your run number input (Format: 00000-00002 or 00002")
                exit(1)

            if rn_end.isdigit()==True:
                eval_nb_max = rn_end
            elif rn_end.isdigit()==False and rn_end == 'MAX':
                eval_nb_max = run_number_max
            else:
                # todo raise exception
                print("Check your run number input (Format: 00000-00002 or 00002")
                exit(1)

        elif run_numbers.find("-") ==-1 and run_numbers.isdigit() == True:
            eval_nb_min = run_numbers
            eval_nb_max = run_numbers


    elif run_numbers == None and run_number_min != None and run_number_max != None:
        eval_nb_min = run_number_min
        eval_nb_max = run_number_max
    else:
        #todo raise execption
        print("Check your run number input (Format: 00000-00002 or 00002")
        exit(1)

    return [eval_nb_min, eval_nb_max]


def eval_run_timestamps(run_timestamps=None, run_timestamp_min=None, run_timestamp_max=None):
    """Function: eval_run_timestamps
    Deals with several ways to hand over the timestamp/run names in aDMIX
    1) <date>_<time> (Single)
    2) <date>_<time>,<date>_<time> (Multiple)
    :param timestamp: A string of timestamps what follows single, multiple criterion
    :return list: A list of timestamps
    """
    eval_ts_min = None
    eval_ts_max = None
    if run_timestamps != None and run_timestamp_min != None and run_timestamp_max != None:

        if run_timestamps.find("-") > 0:

            if run_timestamps.split("-")[0] == 'MIN':
                ts_beg = run_timestamp_min
            else:
                try:
                    ts_beg = string_to_datatime(run_timestamps.split("-")[0])
                except ValueError as e:
                    print("ValueError", e)
                    print("Check your run number input (Format: 180101_1530-180101_1630 or 180101_1530")
                    exit(1)

            if run_timestamps.split("-")[1] == 'MAX':
                ts_end = run_timestamp_max
            else:
                try:
                    ts_end = string_to_datatime(run_timestamps.split("-")[1])
                except ValueError as e:
                    print("ValueError:", e)
                    print("Check your run number input (Format: 180101_1530-180101_1630 or 180101_1530")
                    exit(1)

            if ts_beg < ts_end:
                eval_ts_min = ts_beg
                eval_ts_max = ts_end
            else:
                eval_ts_min = ts_end
                eval_ts_max = ts_beg

        elif run_timestamps.find("-") ==-1 and \
                run_timestamps.split("_")[0].isdigit() and len(run_timestamps.split("_")[0])==6 and \
                run_timestamps.split("_")[1].isdigit() and len(run_timestamps.split("_")[1])==4:

            eval_ts_min = string_to_datatime(run_timestamps)
            eval_ts_max = eval_ts_min
        else:
            print("Check your run number input (Format: 180101_1530-180101_1630 or 180101_1530")
            exit(1)

    elif run_timestamps == None and run_timestamp_min != None and run_timestamp_max != None:
        eval_ts_min = run_timestamp_min
        eval_ts_max = run_timestamp_max
    else:
        #todo raise execption
        print("Check your run number input (Format: 00000-00002 or 00002")
        exit(1)

    return [eval_ts_min, eval_ts_max]


def run_timestamp_converter(timestamp = None):
    """Function: run_timestamp_converter
    Deals with several ways to hand over the timestamp/run names in aDMIX
    1) <date>_<time> (Single)
    2) <date>_<time>,<date>_<time> (Multiple)
    :param timestamp: A string of timestamps what follows single, multiple criterion
    :return list: A list of timestamps
    """
    ts_list = []
    if timestamp != None:
        ts = timestamp.split(",")

        for i_ts in ts:
            i_ts = i_ts.replace(" ", "")

            if i_ts.find("-") >= 0:
                try:
                    beg_i_ts = string_to_datatime(i_ts.split("-")[0])
                    end_i_ts = string_to_datatime(i_ts.split("-")[1])
                except ValueError as e:
                    print("ValueError:", e)
                    print("You need to define a time range such as: 180101_1530-180101_1630")
                    exit(1)
                if end_i_ts > beg_i_ts:
                    ts_list.append( "{beg}-{end}".format(beg=i_ts.split("-")[0], end=i_ts.split("-")[1]) )
            else:
                pass

    else:
        ts_list=None

    return ts_list


def safeformat(str, **kwargs):
    """Function safeformat
    A quick python (3.2+) way to avoid errors with missing keywords in strings.
    Taken from https://stackoverflow.com/questions/17215400/python-format-string-unused-named-arguments
    :param str: The list of strings with brackets
    :param **kwargs: Arguments to fill in the brackets
    :return string: The string with filled brackets but leaves the ones untouched where no information is available
    """

    class SafeDict(dict):
        def __missing__(self, key):
            return '{' + key + '}'
    replacements = SafeDict(**kwargs)
    return str.format_map(replacements)

def read_folder(path):
    """Function: read_folder
    Read a folder from the input path
    :param path: String with valid path information
    :return list: A list of [directory_path, folders, files] found in the path
    """
    ret_folder = []
    ret_dirpath = []
    ret_files = []
    for (dirpath, dirnames, filenames) in os.walk(path):
        ret_dirpath.extend(dirpath)
        ret_folder.extend(dirnames)
        ret_files.extend(filenames)
        break
    return [ret_dirpath, ret_folder, ret_files]

def run_name_converter(run_name=None):
    """Function: run_name_converter
    convert a comma separated list of the --name terminal input into a python  list
    :param run_name: A list of comma separated strings
    :return list: A list of strings
    """

    if run_name != None:
        run_name = run_name.replace(" ", "")
        run_name = run_name.split(",")

    return run_name

def check_valid_timestamp( timestamp=None):
    """Function check_valid_timestamp
    Check for the correct format of string timestamp
    with format <date>_<time>

    :param timestamp: A string timestamp (any format)
    :return bool: True if pattern follows <date>_<time>, otherwise False
    """
    #Check a valid timestamp scheme input
    # <date (YYMMDD)>_<time (hh:mm)>
    ts_valid = False
    if timestamp != None:
        ts = timestamp.split("_")
        if len(ts) == 2 and len(ts[0]) == 6 and len(ts[1]) == 4:
            ts_valid = True

    return ts_valid

def string_to_datatime( time_='700101_0000', pattern='%y%m%d_%H%M'):
    """Function string_to_datetime
    Lazy approach to create a datetime object from a string given a pattern
    :param time: A timestamp string of your choice
    :param pattern: The pattern what applies to the input of time
    :return datetime: A datetime object of the timestampe
    """
    return datetime.datetime.strptime(time_, pattern)


def xenon1t_detector_renamer(input):
    """Function: xenon1t_detector_renamer


    """

    if input.get("detector") == 'muon_veto':
        input['detector'] = 'mv'
    return input




"""
Base class information for different admix daemons
"""
import time
from tqdm import tqdm
from utilix import xent_collection
from admix import logger


class AdmixDaemon:
    query = {}
    projection = {'number': 1}

    # description for tqdm
    desc = "admix loop"

    def __init__(self, db_query=None):

        if db_query is not None:
            # then we loop over all data
            self.query = db_query

        self.collection = xent_collection()

    def data_find(self, limit=0):
        """Uses the db_query to find data that we must do something with"""
        cursor = self.collection.find(self.query, self.projection, limit=limit)
        return cursor

    def do_task(self, rundoc):
        """Define in the children classes"""
        raise NotImplementedError

    def single_loop(self, max_iterations=0, progress_bar=True):
        cursor = list(self.data_find(limit=max_iterations))
        logger.info(f"Running {self.__class__.__name__} on {len(cursor)} entries")
        iterable = cursor
        if progress_bar:
            iterable = tqdm(iterable, desc=self.desc)
        for doc in iterable:
            self.do_task(doc)

    def infinite_loop(self, sleep=600, dt=1, **kwargs):
        try:
            while True:
                self.single_loop(**kwargs)
                # start sleep timer
                total_sleep = sleep
                sleep_left = total_sleep
                while sleep_left > -1:
                    statement = f"Sleeping for {sleep_left} seconds"
                    print(f"{statement:30s}", flush=True, end='\r')
                    time.sleep(dt)
                    sleep_left = sleep_left - dt
        except KeyboardInterrupt:
            print("\nExiting.")



import os.path
from rucio.common.exception import Duplicate, DataIdentifierNotFound
from .rucio import add_scope, list_files, get_rule
from .utils import parse_did, db
from . import clients



def get_default_scope():
    return 'user.' + clients.upload_client.client.account

def preupload(path, rse, did):
    """
    A function supposed to be run before upload to avoid orphan files failing the upload. 
    It does the following
        - It adds the dataset associated to the did we wanted to upload
        - It loops over all local files to be uploaded, so to know their number and their names. 
          For each file, it searches in Rucio catalogue if such a filename is already present. 
          If so, it attaches it to the dataset
        - Finally, it creates a replication rule on the RSE (the RSE is an input parameter of the 
          preupload function, however, it's important that the RSE must be the same chosen by the 
          previous upload attempt). After this latest operation, the did will show up in Rucio.
    """
    if not os.path.isdir(path):
        return

    local_files = os.listdir(path)
    nfiles = len(local_files)
    scope, name = did.split(':')
    try:
        clients.did_client.add_dataset(scope,name)
    except:
        print("DID {0} already exists".format(did))
    for local_file in local_files:
        try:
            clients.did_client.attach_dids(scope,name,[{'scope':scope,'name':local_file}])
        except:
            print("File {0} could not be attached".format(local_file))
    try:
        clients.rule_client.add_replication_rule([{'scope':scope,'name':name}],1,rse)
    except:
        print("The rule for DID {0} already exists".format(did))

# TODO could we make this use multithreading or multiprocessing to speed things up?
def upload(path, rse, 
           register_after_upload=False, verbose=True, did=None, lifetime=None, update_db=False):

    # set scope initially to default one. will overwrite it below if did passed
    scope = get_default_scope()

    if did:
        if ':' in did:
            scope, name = did.split(':')
        else:
            name = did
    else:
        name = os.path.basename(path)
        did = f"{scope}:{name}"

    try:
        add_scope(clients.upload_client.client.account, scope)
    except Duplicate:
        pass

    # compare files attached to this DID to that in path, and only upload missing ones
    if os.path.isdir(path):
        existing_files = []
        local_files = os.listdir(path)
        try:
            existing_files = list_files(did)
        except DataIdentifierNotFound:
            pass

        to_upload = []

        missing_files = set(local_files) - set(existing_files)
        if verbose:
            print(f"Found {len(missing_files)} files to upload:")
            print(missing_files)
            print("------")
        for missing_file in missing_files:
            _path = os.path.join(path, missing_file)
            to_upload.append(dict(path=_path,
                                  rse=rse,
                                  did_scope=scope,
                                  did_name=missing_file,
                                  dataset_scope=scope,
                                  dataset_name=name,
                                  register_after_upload=register_after_upload,
                                  lifetime=lifetime
                                  )
                             )
    else:
        upload_dict = dict(path=path,
                           rse=rse,
                           did_scope=scope,
                           dataset_scope=scope,
                           dataset_name=name,
                           register_after_upload=register_after_upload,
                           lifetime=lifetime
                           )
        to_upload = [upload_dict]

    # get data dict to add to database
    number, dtype, lineage_hash = parse_did(did)
    data_dict = dict(did=did,
                     type=dtype,
                     location=rse,
                     status='transferring',
                     host='rucio-catalogue',
                     meta=dict(),
                     protocol='rucio',
                    )

    if len(to_upload):
        if update_db:
            db.update_data(number, data_dict)
        try:
            clients.upload_client.upload(to_upload)
        except Exception as e:
            if verbose:
                print(f"Upload failed for {path}")
                print(e)
            return did
        # then update db again when complete
        if update_db:
            data_dict['status'] = 'transferred'
            # get files, size etc
            files = list_files(did, verbose=True)
            size = sum([f['bytes'] for f in files]) / 1e6
            data_dict['meta'] = dict(lineage_hash=lineage_hash,
                                     size_mb=size,
                                     file_count=len(files)
                                     )
            db.update_data(number, data_dict)
    else:
        print(f"Nothing to upload at {path}")

    try:
        if not get_rule(did,rse):
            clients.rule_client.add_replication_rule(dids=[{'scope':scope,'name':name}],copies=1,rse_expression=rse)
            print(f"Missing rule has been added")
    except Exception as e:
        if verbose:
            print(f"Insertion of rule failed")
            print(e)
        return did

        
    return did



from rucio.client.client import Client
from rucio.client.replicaclient import ReplicaClient
from rucio.client.accountclient import AccountClient
from rucio.client.rseclient import RSEClient
from rucio.client.downloadclient import DownloadClient
from rucio.client.uploadclient import UploadClient
from rucio.client.ruleclient import RuleClient
from rucio.client.didclient import DIDClient

from . import logger


rucio_client = None
replica_client = None
account_client = None
rse_client = None
download_client = None
upload_client = None
rule_client = None
did_client = None

def _init_clients():
    global rucio_client
    global replica_client
    global account_client
    global rse_client
    global download_client
    global upload_client
    global rule_client
    global did_client

    rucio_client = Client()
    replica_client = ReplicaClient()
    account_client = AccountClient()
    rse_client = RSEClient()
    download_client = DownloadClient(logger=logger)
    upload_client = UploadClient()
    rule_client = RuleClient()
    did_client = DIDClient()

def needs_client(func):
    def wrapped(*args, **kwargs):
        if rucio_client is None:
            _init_clients()
        return func(*args, **kwargs)
    return wrapped



import os
from argparse import ArgumentParser
import admix.helper.helper as helper
from admix import DEFAULT_CONFIG, __version__
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.database import ConnectMongoDB
from admix.utils.naming import make_did
from admix.utils.list_file_replicas import list_file_replicas
from utilix.config import Config
import utilix
from bson.json_util import dumps
from datetime import timezone, datetime, timedelta
import pymongo

def showrun(arg_number,arg_to,arg_dtypes,arg_compact,arg_dumpjson,arg_status,arg_latest):

    #Define data types
    NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
    RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
    RECORDS_DTYPES = helper.get_hostconfig()['records_types']

    #Get other parameters
    DATADIR = helper.get_hostconfig()['path_data_to_upload']
    periodic_check = helper.get_hostconfig()['upload_periodic_check']
    RSES = helper.get_hostconfig()['rses']

    minimum_number_acceptable_rses = 2
    minimum_deltadays_allowed = 3

    # Storing some backup hashes in case DID information is not available
    bkp_hashes = { 'raw_records':'rfzvpzj4mf', 'raw_records_he':'rfzvpzj4mf', 'raw_records_mv':'rfzvpzj4mf', 'raw_records_aqmon':'rfzvpzj4mf', 'records':'56ausr64s7', 'lone_hits':'b7dgmtzaef' }

    context = 'xenonnt_online'

    #Init the runDB
    db = ConnectMongoDB()

    #Init Rucio for later uploads and handling:
    rc = RucioSummoner(helper.get_hostconfig("rucio_backend"))
    rc.SetRucioAccount(helper.get_hostconfig('rucio_account'))
    rc.SetConfigPath(helper.get_hostconfig("rucio_cli"))
    rc.SetProxyTicket(helper.get_hostconfig('rucio_x509'))
    rc.SetHost(helper.get_hostconfig('host'))
    rc.ConfigHost()
    rc.SetProxyTicket("rucio_x509")


    data_types = RAW_RECORDS_DTYPES + RECORDS_DTYPES + NORECORDS_DTYPES

    if arg_number == -1 and arg_latest == 0:
        arg_latest = 5
        arg_compact = True

    if arg_latest > 0:
        cursor = db.db.find({}).sort('number',pymongo.DESCENDING).limit(1)
        cursor = list(cursor)
        arg_to = cursor[0]['number']
        arg_number = arg_to - arg_latest + 1
        print('Processing latest {0} runs'.format(arg_latest))            

    if arg_to>arg_number:
        cursor = db.db.find({
            'number': {'$gte': arg_number, '$lte': arg_to}
        }).sort('number',pymongo.ASCENDING)
        print('Runs that will be processed are from {0} to {1}'.format(arg_number,arg_to))
    else:
        cursor = db.db.find({
            'number': arg_number
        })
        print('Run that will be processed is {0}'.format(arg_number))
    cursor = list(cursor)

    # Runs over all listed runs
    for run in cursor:

        print("")

        # Gets run number
        number = run['number']
        print('Run: {0}'.format(number))

        # Gets the status
        if 'status' in run:
            print('Status: {0}'.format(run['status']))
        else:
            print('Status: {0}'.format('Not available'))

        if arg_status:
            continue

        # Extracts the correct Event Builder machine who processed this run
        # Then also the bootstrax state and, in case it was abandoned, the reason
        if 'bootstrax' in run:
            bootstrax = run['bootstrax']
            eb = bootstrax['host'].split('.')[0]
            print('Processed by: {0}'.format(eb))
            if 'state' in bootstrax:
                print('Bootstrax state: {0}'.format(bootstrax['state']))
                if bootstrax['state'] == 'abandoned':
                    if 'reason' in bootstrax:
                        print('Reason: {0}'.format(bootstrax['reason']))
        else:
            print('Not processed')

        # Gets the date
        if 'start' in run:
            start_time = run['start'].replace(tzinfo=timezone.utc)
            print("Date: ",start_time.astimezone(tz=None))
            
            # Calculates the duration
            if 'end' in run:
                if run['end'] is not None:
                    end_time = run['end'].replace(tzinfo=timezone.utc)
                    duration = end_time-start_time
                    print("Duration: ",duration)
                else:
                    print("Duration: ","unknown")
        
            # Prints if run is still enough recent (three days from now)
            now_time = datetime.now().replace(tzinfo=timezone.utc)
            delta_time = now_time-start_time
            if delta_time < timedelta(days=minimum_deltadays_allowed):
                print("Less than {0} days old".format(minimum_deltadays_allowed))
        else:
            print("Warning : no time info available")


        # Gets the comments
        if 'comments' in run:
            if len(run['comments'])>0:
                last_comment = run['comments'][-1]
                print("Latest comment ({0}): {1}".format(last_comment['user'],last_comment['comment']))
        

        # Dumps the entire rundoc under json format
        if arg_dumpjson:
            print(dumps(run, indent=4))

        if arg_compact:
            continue

        # Merges data and deleted_data
        if 'deleted_data' in run:
            data = run['data'] + run['deleted_data']
        else:
            data = run['data']
            
        # Check is there are more instances in more EventBuilders
        extra_ebs = set()
        for d in data:
            if 'eb' in d['host'] and eb not in d['host']: 
                extra_ebs.add(d['host'].split('.')[0])
        if len(extra_ebs)>0:
            print('\t\t Warning : The run has been processed by more than one EventBuilder: {0}'.format(extra_ebs))

        # Runs over all data types to be monitored
        for dtype in data_types:

            if len(arg_dtypes)>0:
                if dtype not in arg_dtypes:
                    continue

            # Data type name
            print('{0}'.format(dtype))

            # Take the official number of files accordingto run DB
            # and the eb status
            Nfiles = -1
            ebstatus = ""
            for d in data:
                if d['type'] == dtype and eb in d['host']:
                    if 'file_count' in d:
                        Nfiles = d['file_count']
                    if 'status' in d:
                        ebstatus = d['status']
            if Nfiles == -1:
                print('\t Number of files: missing in DB')
            else:
                print('\t Number of files: {0}'.format(Nfiles))

            if ebstatus != "":
                print('\t EB status: {0}'.format(ebstatus))
            else:
                print('\t EB status: not available')


            # Check if data are still in the data list and not in deleted_data
            DB_InEB = False
            for d in run['data']:
                if d['type'] == dtype and eb in d['host']:
                    DB_InEB = True
            DB_NotInEB = False
            if 'deleted_data' in run:
                for d in run['deleted_data']:
                    if d['type'] == dtype and eb in d['host']:
                        DB_NotInEB = True
            if DB_InEB and not DB_NotInEB:
                print('\t DB : still in EB')
            if not DB_InEB and DB_NotInEB:
                print('\t DB : deleted from EB')
            if DB_InEB and DB_NotInEB:
                print('\t\t Incoherency in DB: it is both in data list and in deleted_data list')
            #if (DB_InEB and DB_NotInEB) or (not DB_InEB and not DB_NotInEB):
            #  print('\t\t incoherency in DB: it is neither in data list nor in deleted_data list')

            # Check if data are still in the EB disks without using the DB
            upload_path = ""
            for d in run['data']:
                if d['type'] == dtype and eb in d['host']:
                    file = d['location'].split('/')[-1]
                    upload_path = os.path.join(DATADIR, eb, file) 
            path_exists = os.path.exists(upload_path)
            if upload_path != "" and path_exists:
                path, dirs, files = next(os.walk(upload_path))
                print('\t Disk: still in EB disk and with',len(files),'files')
            else:
                print('\t Disk: not in EB disk')
            if DB_InEB and not path_exists:
                print('\t\t Incoherency in DB and disk: it is in DB data list but it is not in the disk')
            if DB_NotInEB and path_exists:
                print('\t\t Incoherency in DB and disk: it is in DB deleted_data list but it is still in the disk')

            # The list of DIDs (usually just one)
            dids = set()
            for d in data:
                if d['type'] == dtype and d['host'] == 'rucio-catalogue':
                    if 'did' in d:
                        dids.add(d['did'])
            print('\t DID:', dids)

            # Check the presence in each available RSE
            Nrses = 0
            for rse in RSES:
                is_in_rse = False
                for d in run['data']:
                    if d['type'] == dtype and rse in d['location']:
                        if 'status' in d:
                            status = d['status']
                        else:
                            status = 'Not available'
                        if 'did' in d:
                            hash = d['did'].split('-')[-1]
                            did = d['did']
                        else:
                            print('\t\t Warning : DID information is absent in DB data list (old admix version). Using standard hashes for RSEs')
                            #hash = bkp_hashes.get(dtype)
                            #hash = utilix.db.get_hash(context, dtype)
                            hash = db.GetHashByContext(context,dtype)
                            did = make_did(number, dtype, hash)
                        rucio_rule = rc.GetRule(upload_structure=did, rse=rse)
                        files = list_file_replicas(number, dtype, hash, rse)
                        if rucio_rule['exists']:
                            print('\t', rse+': DB Yes, Status',status,', Rucio Yes, State',rucio_rule['state'],",",len(files), 'files')
                            if len(files) < Nfiles and rucio_rule['state']!="REPLICATING":
                                print('\t\t Warning : Wrong number of files in Rucio!!!')
                        else:
                            print('\t', rse+': DB Yes, Status',status,', Rucio No')
                        # print(files)
                        is_in_rse = True
                        Nrses += 1
                if not is_in_rse:
#                    print('\t\t Warning : data information is absent in DB data list. Trying using standard hashes to query Rucio')
#                    hash = bkp_hashes.get(dtype)
                    #hash = utilix.db.get_hash(context, dtype)
                    hash = db.GetHashByContext(context,dtype)
                    did = make_did(number, dtype, hash)
                    print('\t Guessed DID:', did)
                    rucio_rule = rc.GetRule(upload_structure=did, rse=rse)
                    files = list_file_replicas(number, dtype, hash, rse)
                    if rucio_rule['exists']:
                        print('\t', rse+': DB No, Rucio Yes, State',rucio_rule['state'],",",len(files), 'files')
                        if len(files) < Nfiles and rucio_rule['state']!="REPLICATING":
                            print('\t\t Warning : Wrong number of files in Rucio!!!')
                    else:
                        print('\t', rse+': DB No, Rucio No')
            print('\t Number of sites: ', Nrses)



    

def main():
    parser = ArgumentParser("admix-showrun")

    config = Config()

    parser.add_argument("number", type=int, nargs='?', help="Run number to show", default=-1)
    parser.add_argument("--dtypes", nargs="*", help="Restricts infos on the given data types")
    parser.add_argument("--to", type=int, help="Shows runs from the run number up to this value", default=0)
    parser.add_argument("--compact", help="Just list few DB infos as run number, status, date, comments", action='store_true')
    parser.add_argument("--status", help="Just list the run name and its global status", action='store_true')
    parser.add_argument("--json", help="Dumps the whole DB rundoc in pretty style", action='store_true')
    parser.add_argument("--latest", type=int, help="Shows latest runs", default=0)

    args = parser.parse_args()

    if args.dtypes:
        dtypes = args.dtypes
    else:
        dtypes = []

    helper.make_global("admix_config", os.path.abspath(config.get('Admix','config_file')))

    try:
        showrun(args.number,args.to,dtypes,args.compact,args.json,args.status,args.latest)
    except KeyboardInterrupt:
        return 0






import os
import time
import shutil
from argparse import ArgumentParser
import admix.helper.helper as helper
from admix import DEFAULT_CONFIG, __version__
from admix.interfaces.rucio_summoner import RucioSummoner
from admix.interfaces.database import ConnectMongoDB
from admix.utils.naming import make_did
from admix.utils.list_file_replicas import list_file_replicas
from utilix.config import Config
import utilix
from bson.json_util import dumps
from datetime import timezone, datetime, timedelta
import pymongo

class Fix():

    def __init__(self):

        #Take all data types categories
        self.NORECORDS_DTYPES = helper.get_hostconfig()['norecords_types']
        self.RAW_RECORDS_DTYPES = helper.get_hostconfig()['raw_records_types']
        self.RECORDS_DTYPES = helper.get_hostconfig()['records_types']

        #Choose which data type you want to treat
        self.DTYPES = self.NORECORDS_DTYPES + self.RECORDS_DTYPES + self.RAW_RECORDS_DTYPES
        
        #Take the list of all XENON RSEs
        self.RSES = helper.get_hostconfig()['rses']

        #Take the RSE that is used to perform the upload
        self.UPLOAD_TO = helper.get_hostconfig()['upload_to']

        #Init the runDB
        self.db = ConnectMongoDB()

        #Init Rucio for later uploads and handling:
        self.rc = RucioSummoner()

        #Rucio Rule assignment priority
        self.priority = 3

    def reset_upload(self,did):

        hash = did.split('-')[-1]
        dtype = did.split('-')[0].split(':')[-1]
        number = int(did.split(':')[0].split('_')[-1])

        print("Resetting the upload associated to the DID: {0}".format(did))
        print("Run number: {0}".format(number))
        print("Data type: {0}".format(dtype))
        print("Hash: {0}".format(hash))

        run = self.db.db.find_one({'number' : number})

        # Gets the status
        if 'status' in run:
            print('Run status: {0}'.format(run['status']))
        else:
            print('Run status: {0}'.format('Not available'))

        # Extracts the correct Event Builder machine who processed this run
        # Then also the bootstrax state and, in case it was abandoned, the reason
        if 'bootstrax' in run:
            bootstrax = run['bootstrax']
            eb = bootstrax['host'].split('.')[0]
        else:
            print('Not processed')
            return(0)
                

        # Get the EB datum and its status
        ebstatus = ""
        datum = None
        for d in run['data']:
            if d['type'] == dtype and eb in d['host']:
                datum = d
                if 'status' in d:
                    ebstatus = d['status']

        if datum is None:
            print('There is no EB datum. No reset is possible')
            return(0)

        if ebstatus != "":
            print('EB status: {0}'.format(ebstatus))
        else:
            print('EB status: not available')


        # Step zero (normally not needed): change the run status to "transferring"
        #    self.db.db.find_one_and_update({'number':number},{'$set':{"status": "transferring"}})


        # First action: remove the files stored in datamanager
        files = list_file_replicas(number, dtype, hash, self.UPLOAD_TO)
        print("Deleting rucio data in datamanager disk. Deleting",len(files),"files")
        for file in files:
            try:
                os.remove(file)
            except:
                print("File: {0} not found".format(file))



        # Second action: remove the LNGS Rucio rule
        deleted_any_rule = False
        for rse in self.RSES:
            rucio_rule = self.rc.GetRule(upload_structure=did, rse=rse)
            if rucio_rule['exists']:
                print("Deleting rucio rule = ", rucio_rule['id'], "from RSE = ",rse)
                self.rc.DeleteRule(rucio_rule['id'])
                deleted_any_rule = True
        
        # If some rule has been deleted, wait for 1 hour (plus 5 minutes of margin)
        if deleted_any_rule:
            delay = 3600+60*5
            print("We have to wait for {0} seconds before proceeding to the next step".format(delay))
            time.sleep(delay)
        else:
            print("There is no rule to delete")



        # Third action: set the EB status as 'eb_ready_to_upload' 
        self.db.db.find_one_and_update({'_id': run['_id'],'data': {'$elemMatch': datum}},
                                  {'$set': {'data.$.status': 'eb_ready_to_upload'}})
        print("EB status changed to eb_ready_to_upload")



        # Reload the run
        run = self.db.db.find_one({'number' : number})

        # Gets the status
        if 'status' in run:
            print('New run status: {0}'.format(run['status']))
        else:
            print('Ru status: {0}'.format('Not available'))

        # Get the EB datum and its status
        ebstatus = ""
        datum = None
        for d in run['data']:
            if d['type'] == dtype and eb in d['host']:
                datum = d
                if 'status' in d:
                    ebstatus = d['status']

        # Prints the eb status as a confirmation of the performed change 
        if ebstatus != "":
            print('New EB status: {0}'.format(ebstatus))
        else:
            print('New EB status: not available')


    def add_rule(self,did,from_rse,to_rse):

        hash = did.split('-')[-1]
        dtype = did.split('-')[0].split(':')[-1]
        number = int(did.split(':')[0].split('_')[-1])

        print("Adding a new rule {0} from {1} to {2}".format(did,from_rse,to_rse))
        print("Run number: {0}".format(number))
        print("Data type: {0}".format(dtype))
        print("Hash: {0}".format(hash))

        run = self.db.db.find_one({'number' : number})

        # Gets the status
        if 'status' in run:
            print('Run status: {0}'.format(run['status']))
        else:
            print('Run status: {0}'.format('Not available'))

        #Checks if the datum of the sender exists in the DB
        datum = None
        for d in run['data']:
            if d['type'] == dtype and d['host'] == 'rucio-catalogue' and d['location'] == from_rse:
                datum = d
                break
        if datum is None:
            print('The datum concerning data type {0} and site {1} is missing in the DB. Forced to stop'.format(dtype,from_rse))
            return(0)


        # Checks the rule status of the sender RSE
        rucio_rule = self.rc.GetRule(upload_structure=did, rse=from_rse)
        if rucio_rule['state'] != 'OK' and rucio_rule['state'] != 'REPLICATING':
            print('The rule in {0} is neither OK nor REPLICATING. Forced to stop'.format(from_rse))
            return(0)

        # set the new rule
        if not self.skip_rucio:
            print("Adding the Rucio rule")
            self.rc.AddConditionalRule(did, from_rse, to_rse, lifetime=None, priority=self.priority)
        else:
            print("Rucio rule is not added")
        rucio_rule = self.rc.GetRule(did, rse=to_rse)

        # Update run status
        self.db.db.find_one_and_update({'number': number},{'$set': {'status': 'transferring'}})

        # Add a new datum in the run document
        updated_fields = {'host': "rucio-catalogue",
                          'type': dtype,
                          'location': to_rse,
                          'lifetime': rucio_rule['expires'],
                          'status': 'transferring',
                          'did': did,
                          'protocol': 'rucio'
                }
        data_dict = datum.copy()
        data_dict.update(updated_fields)
        self.db.AddDatafield(run['_id'], data_dict)

        print("Done.")



    def delete_rule(self,did,rse):

        hash = did.split('-')[-1]
        dtype = did.split('-')[0].split(':')[-1]
        number = int(did.split(':')[0].split('_')[-1])

        print("Deleting the rule {0} from {1}".format(did,rse))
        print("Run number: {0}".format(number))
        print("Data type: {0}".format(dtype))
        print("Hash: {0}".format(hash))

        run = self.db.db.find_one({'number' : number})

        #Checks if the datum exists in the DB
        datum = None
        for d in run['data']:
            if d['type'] == dtype and d['host'] == 'rucio-catalogue' and d['location'] == rse:
                datum = d
                break
            
        #Delete the datum
        if datum is not None:
            self.db.RemoveDatafield(run['_id'],datum)
            print("Datum deleted in DB.")
        else:
            print('There is no datum to delete')

        #Get the rule of a given DID
        rucio_rule = self.rc.GetRule(upload_structure=did, rse=rse)

        #Delete the rule
        if rucio_rule['exists']:
            self.rc.DeleteRule(rucio_rule['id'])
            print("Rucio rule deleted.")
        else:
            print('There is no Rucio rule to delete')

        print("Done.")



    def delete_db_datum(self,did,site):

        hash = did.split('-')[-1]
        dtype = did.split('-')[0].split(':')[-1]
        number = int(did.split(':')[0].split('_')[-1])

        print("Removing the datum from DB for the DID: {0} and from the site {1}".format(did,site))
        print("Run number: {0}".format(number))
        print("Data type: {0}".format(dtype))
        print("Hash: {0}".format(hash))
        print("Site: {0}".format(site))

        run = self.db.db.find_one({'number' : number})

        # Get the EB datum and its status
        datum = None
        for d in run['data']:
            if 'eb' in site:
                if d['type'] == dtype and site in d['host'] and 'xenon.local' in d['host']:
                    datum = d
                    break
            else:
                if d['type'] == dtype and d['host']=='rucio-catalogue' and d['location']==site:
                    datum = d
                    break
                

        if datum is not None:
            self.db.RemoveDatafield(run['_id'],datum)
            print("Done.")
        else:
            print('There is no datum. Nothing has been deleted')


    def set_run_status(self,number,status):

        number = int(number)

        print("Setting the status of run {0} to the value {1}".format(number,status))

        run = self.db.db.find_one({'number' : number})
        print("status before = ",run['status'])

        self.db.db.find_one_and_update({'_id': run['_id']},{'$set':{"status": status}})

        run = self.db.db.find_one({'number' : number})
        print("status after = ",run['status'])


    def set_eb_status(self,did,status):

        print("Setting the EB status of DID {0} to the value {1}".format(did,status))

        hash = did.split('-')[-1]
        dtype = did.split('-')[0].split(':')[-1]
        number = int(did.split(':')[0].split('_')[-1])

        print("Run number: {0}".format(number))
        print("Data type: {0}".format(dtype))
        print("Hash: {0}".format(hash))

        run = self.db.db.find_one({'number' : number})

        # Extracts the correct Event Builder machine who processed this run
        # Then also the bootstrax state and, in case it was abandoned, the reason
        if 'bootstrax' in run:
            bootstrax = run['bootstrax']
            eb = bootstrax['host'].split('.')[0]
        else:
            print('Not processed')
            return(0)

        # Get the EB datum and its status
        ebstatus = ""
        datum = None
        for d in run['data']:
            if d['type'] == dtype and eb in d['host']:
                datum = d
                if 'status' in d:
                    ebstatus = d['status']

        if datum is None:
            print('There is no EB datum.')
            return(0)

        if ebstatus != "":
            print("EB status before = ",ebstatus)
        else:
            print("EB status absent before")

        #Set the aimed value
        self.db.db.find_one_and_update({'_id': run['_id'],'data': {'$elemMatch': datum}},
                                       {'$set': {'data.$.status': status}})

        run = self.db.db.find_one({'number' : number})

        # Get the EB datum and its status
        ebstatus = ""
        datum = None
        for d in run['data']:
            if d['type'] == dtype and eb in d['host']:
                datum = d
                if 'status' in d:
                    ebstatus = d['status']

        print("EB status after = ",ebstatus)



        

    def __del__(self):
        pass




    

def main():
    parser = ArgumentParser("admix-fix")

    config = Config()

#    parser.add_argument("--number", type=int, help="Run number to fix", default=-1)
#    parser.add_argument("--dtype", help="Data type to fix", default="")
#    parser.add_argument("--did", help="DID to fix")
#    parser.add_argument("--action", help="Which action you want to take")
#    parser.add_argument("--fromrse", help="From which RSE you want to copy data")
#    parser.add_argument("--torse", help="To which RSE you want to copy data")

    parser.add_argument("--reset_upload", nargs=1, help="Deletes everything related a given DID, exept data in EB. The deletion includes the entries in the Rucio catalogue and the related data in the DB rundoc. This is ideal if you want to retry an upload that failed", metavar=('DID'))
    parser.add_argument("--add_rule", nargs=3, help="Add a new replication rule of a given DID from one RSE to another one. The rundoc in DB is updated with a new datum as well", metavar=('DID','FROM_RSE','TO_RSE'))
    parser.add_argument("--delete_rule", nargs=2, help="Delete a replication rule of a given DID from one RSE. The rundoc in DB is deleted as well", metavar=('DID','RSE'))
    parser.add_argument("--delete_db_datum", nargs=2, help="Deletes the db datum corresponding to a given DID. The SITE can be either a specific EB machine (ex: eb1) or a specific RSE", metavar=('DID','SITE'))

    parser.add_argument("--set_run_status", nargs=2, help="Set the run status to a given NAME (typical case is to set it to eb_ready_to_upload)", metavar=('RUN_NUMBER','STATUS_NAME'))
    parser.add_argument("--set_eb_status", nargs=2, help="Set the EB status of a given DID to a given NAME", metavar=('DID','STATUS_NAME'))

    parser.add_argument("--priority", type=int, help="Priority to assign to Rucio rules (default: %(default)s)", default=3)
    parser.add_argument("--skip_rucio", help="Add this flag in context of add_rule in case you just want to update DB since Rucio rule exists already", action='store_true')

    args = parser.parse_args()


    helper.make_global("admix_config", os.path.abspath(config.get('Admix','config_file')))

    fix = Fix()

    fix.skip_rucio = args.skip_rucio
    fix.priority = args.priority

    try:
        if args.reset_upload:
            fix.reset_upload(args.reset_upload[0])
        if args.add_rule:
            fix.add_rule(args.add_rule[0],args.add_rule[1],args.add_rule[2])
        if args.delete_rule:
            fix.delete_rule(args.delete_rule[0],args.delete_rule[1])
        if args.delete_db_datum:
            fix.delete_db_datum(args.delete_db_datum[0],args.delete_db_datum[1])

        if args.set_run_status:
            fix.set_run_status(args.set_run_status[0],args.set_run_status[1])
        if args.set_eb_status:
            fix.set_eb_status(args.set_eb_status[0],args.set_eb_status[1])

#        if args.action == "reset_upload" and args.did:
#            fix.reset_upload(args.did)
#        if args.action == "add_rule" and args.did and args.fromrse and args.torse:
#            fix.add_rule(args.did,args.fromrse,args.torse)
        print("")
    except KeyboardInterrupt:
        return 0




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

# Get requirements from requirements.txt, stripping the version tags
with open('requirements.txt') as f:
    requirements = [r.split('/')[-1] if r.startswith('git+') else r for r in f.read().splitlines()]


setup(
    name='xe-admix',
    version='1.0.15',
    description="advanced Data Management In Xenon (aDMIX)",
    long_description=readme + '\n\n' + history,
    url='https://github.com/XENON1T/admix',
    install_requires=requirements,
    scripts=['bin/admix-download'],
    packages=find_packages(),
    license="BSD license",
    zip_safe=False,
    keywords='admix',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6'
    ],
    test_suite='tests',
)



"""


"""

import os
import hashlib
from admix import logger
from admix.interfaces.rucio_api import RucioAPI


class RucioSummoner():
    def __init__(self, rucio_backend="API"):
        # Do imports outside inside the init so as not to raise errors
        # in rucio on travis tests
        from admix.interfaces.rucio_api import RucioAPI

        self.rucio_backend=rucio_backend
        self.rucio_account = os.environ.get("RUCIO_ACCOUNT")
        self._rucio = None
        if self.rucio_backend=="API":
            self._rucio = RucioAPI()
        else:
            print(f"You chose {self.rucio_backend} as Rucio backend which does not exists")
            print("Fix this!")
            exit()

    #Here comes the frontend configuration part:
    def SetRucioAccount(self, rucio_account):
        self.rucio_account = rucio_account
        self._rucio.SetRucioAccount(rucio_account)

    def SetConfigPath(self, config_path):
        self._rucio.SetConfigPath(config_path)

    def SetProxyTicket(self, proxy_path):
        self._rucio.SetProxyTicket(proxy_path)

    def SetHost(self, hostname):
        self._rucio.SetHost(hostname)

    def ConfigHost(self):
        self._rucio.ConfigHost()

    def Whoami(self):
        """Rucio Summoner:Whoami
           Results a dictionary to identify the current
           Rucio user and credentials.

        :return A dictionary with Rucio whoami information
        """

        return self._rucio.Whoami()

    def Alive(self):
        """Function: Alive
        Simple print statement to test Rucio setup
        """
        whoami = self._rucio.Whoami()

        print("Rucio ")
        print("Rucio Whoami()")
        for ikey, ival in whoami.items():
            print(ikey, "\t \t", ival)
        print()
        print("Rucio alive")

    def _md5_hash(self, string):
        """Function: _md5_hash(...)

        Calculate a md5 hash from a string

        :param string: A string
        :return result: A md5 checksum of the input string
        """
        return hashlib.md5(string.encode('utf-8')).hexdigest()

    def _VerifyStructure(self, upload_structure=None, level=-1):
        """The Rucio summoner is able to deal with
        two kinds of valid input arguments. To avoid
        a break in the command chain we verify the
        structure here first and prepare further steps.
        The two valid input arguments are:
        - A Rucio scope:name structure (DID) which is encoded
          by a string
        - A stacked container-dataset-file structure which
          is encoded in a dictionary

        :param: upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param: level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                which the 'did' is chosen from.
        :return: (val_scope, val_dname): The extracted Rucio DID to which certain operations are applied.
        """

        val_scope = None
        val_dname = None

        if isinstance(upload_structure, str):
            try:
                val_scope = upload_structure.split(":")[0]
                val_dname = upload_structure.split(":")[1]
            except IndexError as e:
                print("Function _VerifyStructure for Rucio DID input: IndexError")
                print("Message:", e)
                #exit(1)
        elif isinstance(upload_structure, dict):
            #you can not sort keys if they are not in the dictionary:
            #leave like it is
            sorted_keys = [key for key in sorted(upload_structure.keys())]
            try:
                val_scope = upload_structure[sorted_keys[level]]['did'].split(":")[0]
                val_dname = upload_structure[sorted_keys[level]]['did'].split(":")[1]
            except IndexError as e:
                print("Function _VerifyStructure for Rucio template input: IndexError")
                print("Message:", e)
                #exit(1)

        return (val_scope, val_dname)

    def AddRule(self, did, rse, lifetime=None, protocol='rucio-catalogue', priority=3):
        """Add rules for a Rucio DID or dictionary template.

        :param: did: Rucio DID form of "scope:name"
        :param: rse: An existing Rucio storage element (RSE)
        :param: lifetime: Choose a lifetime of the transfer rule in seconds or None
        :param: protocol: Should always be 'rucio-catalogue'?
        :return:
        """

        # analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(did)

        # Get current rules for this did
        rules = self._rucio.ListDidRules(val_scope, val_dname)
        current_rses = [r['rse_expression'] for r in rules]

        # if a rule already exists, exit
        if rse in current_rses:
            print("There already exists a rule for DID %s at RSE %s" % (did, rse))
            return 1

        # add rule
        did_dict= {}
        did_dict['scope'] = val_scope
        did_dict['name']  = val_dname

        self._rucio.AddRule( [did_dict],
                             copies=1,
                             rse_expression=rse,
                             lifetime=lifetime,
                             priority=priority)
        return 0


    def AddConditionalRule(self, did, from_rse, to_rse, lifetime=None, protocol='rucio-catalogue', priority=3):
        """Add rules for a Rucio DID or dictionary template.

        :param: did: Rucio DID form of "scope:name"
        :param: rse: An existing Rucio storage element (RSE)
        :param: lifetime: Choose a lifetime of the transfer rule in seconds or None
        :param: protocol: Should always be 'rucio-catalogue'?
        :return:
        """

        # analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(did)

        # Get current rules for this did
        rules = self._rucio.ListDidRules(val_scope, val_dname)
        current_rses = [r['rse_expression'] for r in rules]

        # if a rule already exists, exit
        if to_rse in current_rses:
            print("There already exists a rule for DID %s at RSE %s" % (did, to_rse))
            return 1

        # add rule
        did_dict= {}
        did_dict['scope'] = val_scope
        did_dict['name']  = val_dname

        self._rucio.AddRule( [did_dict],
                             copies=1,
                             rse_expression=to_rse,
                             source_replica_expression=from_rse,
                             lifetime=lifetime,
                             priority=priority)
        return 0


    def UpdateRules(self, upload_structure=None, rse_rules=None, level=-1):
        """Update existing rules for a Rucio DID or dictionary template.

        :param: upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param: rse_rules: A list of strings which follow a certain template of ["{protocol}:{rse}:{lifetime}",...]
                           With:
                           protocol: rucio-catalogue
                           rse: An existing Rucio storage element (RSE)
                           lifetime: Choose a lifetime of the transfer rule in seconds or None
        :param: level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                       which the 'did' is chosen from.
        :return result: A dictionary with Rucio Storage Elements (RSE) as keys. The value is another dictionary
                        with keys 'result' (0 on success, 1 on failure) and lifetime ( an integer > 0)
        """

        #analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(upload_structure, level)

        #Get Rule IDs:
        r_rules = self._rucio.ListDidRules(val_scope, val_dname)
        r_rse_ids = {}
        for i_rule in r_rules:
            r_rse = i_rule['rse_expression']
            r_rse_ids[r_rse] = i_rule['id']

        result = {}
        for i_rule in rse_rules:
            g_ptr = i_rule.split(":")[0]
            g_rse = i_rule.split(":")[1]
            g_rlt = i_rule.split(":")[2]
            if g_rlt == "None":
                g_rlt = None
            else:
                g_rlt = int(g_rlt)
            if g_rse not in list(r_rse_ids.keys()):
                continue

            options = {}
            options['lifetime'] = g_rlt
            r = self._rucio.UpdateRule(r_rse_ids[g_rse], options )

            result[g_rse] = {}
            result[g_rse]['result'] = r
            result[g_rse]['lifetime'] = g_rlt

        return result

    def DeleteRule(self, rule_id):
        self._rucio.DeleteRule(rule_id)

    def _rule_status_dictionary(self):
        """This dictionary defines the full set of rule information
        what is returned from Rucio and dedicated to further usage.
        Add information carefully if you need to. Removing anything from
        this dictionary breaks aDMIX."""
        rule = {}
        rule['rse'] = None
        rule['exists'] = False
        rule['state'] = "Unkown"
        rule['cnt_ok'] = 0
        rule['cnt_repl'] = 0
        rule['cnt_stuck'] = 0
        rule['id'] = None
        rule['expires'] = None
        return rule

    def ListDidRules(self, upload_structure=None, level=-1):
        """List existing rules for a Rucio DID or dictionary template.

        :param: upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param: level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                which the 'did' is chosen from.
        :return: A list of Rucio transfer rules with additional rule information. Each list element stands for a
                 Rucio Storage Element (RSE). If no rule exists it returns an empty list

        """

        #analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(upload_structure, level)

        r_rules = self._rucio.ListDidRules(val_scope, val_dname)

        return r_rules

    def ListFileReplicas(self, upload_structure=None, rse=None, level=-1, localpath=False):
        """Function: ListFileReplicas(...)

        List all your file replicas which are attached to a dataset or container.

        Hint: List of RSE wide file replicas (local path) was not available in Rucio 1.19.

        :param upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param rse: A valid Rucio Storage Element (RSE) of the current Rucio setting.
        :param level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                      which the 'did' is chosen from.
        :return result: Dictionary which with key->value ordering follows:\n
                         - key: filename of the attached file\n
                         - value: The local file location for the selected RSE\n
                        Otherwise: {}

        """

        # Prepare a return dictionary:
        result = {}

        #Begin to list all files which are in the upload_structure:
        file_list = [ i_file['name'] for i_file in self.ListFiles(upload_structure, level)]

        #analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(upload_structure, level)

        #Extract all valid RSEs in the current Rucio setup and return [] if rse not valid
        list_rse = [ i_rse['rse'] for i_rse in list(self._rucio.ListRSEs())]
        if rse not in list_rse:
            return result

        # get RSE overview if RSE exists
        rse_overview = self._rucio.GetRSE(rse)

        # check if the given RSE is tape (true for yes)
        _istape = False
        if rse_overview.get('rse_type') == "DISK":
            _istape=False
        else:
            _istape=True

        rse_hostname = rse_overview['protocols'][0]['hostname']
        rse_prefix = rse_overview['protocols'][0]['prefix']
        rse_port = rse_overview['protocols'][0]['port']
        rse_scheme = rse_overview['protocols'][0]['scheme']

        # Prepare basic scheme for RSE locations:
        lfn = None
        lfn_disk = "{protocol}://{hostname}:{port}{prefix}/{scope}/{h1}/{h2}/{fname}"
        lfn_local_disk = "{prefix}/{scope}/{h1}/{h2}/{fname}"
        lfn_tape = "{protocol}://{hostname}:{port}{prefix}/{scope}/{fname}"


        #Start to fill the dictionary according to istape==True or False
        if _istape==False:
            # assume a disk storage (deterministic paths):
            for i_filename in file_list:
                # Calculate the deterministic hash sums:
                rucio_did = "{scope}:{name}".format(scope=val_scope,
                                                    name=i_filename)
                t1 = self._md5_hash(rucio_did)[0:2]
                t2 = self._md5_hash(rucio_did)[2:4]

                if localpath:
                    lfn = lfn_local_disk.format(prefix=rse_prefix,
                                          scope=val_scope,
                                          h1=t1,
                                          h2=t2,
                                          fname=i_filename)
                    result[i_filename] = lfn
                else:
                    lfn = lfn_disk.format(protocol=rse_scheme,
                                          hostname=rse_hostname,
                                          port=rse_port,
                                          prefix=rse_prefix,
                                          scope=val_scope,
                                          h1=t1,
                                          h2=t2,
                                          fname=i_filename)
                    result[i_filename] = lfn


        else:
            #assume tape storage (non-deterministic paths)
            for i_filename in file_list:
                lfn = lfn_tape.format(protocol=rse_scheme,
                                      hostname=rse_hostname,
                                      port=rse_port,
                                      prefix=rse_prefix,
                                      scope=val_scope,
                                      fname=i_filename)
                result[i_filename] = lfn

        return result

    def ListFiles(self, upload_structure=None, long=None, level=-1):
        """List existing files for a Rucio DID or dictionary template.

        :param: upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param: level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                which the 'did' is chosen from.
        :param long: Define another output (Check the Rucio tutorials for it)
        :return result: A list of files, otherwise []
        """

        #analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(upload_structure, level)

        result = []
        result = self._rucio.ListFiles(val_scope, val_dname, long=long)

        return result

    def ListDids(self,scope, filters, type='collection', long=False, recursive=False):
        """
        List all data identifiers in a scope which match a given pattern. Check Rucio github page for details

        :param scope: The valid string which follows the Rucio scope name.
        :param filters: A dictionary of key/value pairs like {'name': 'file_name','rse-expression': 'tier0'}.
        :param type: The type of the did: 'all'(container, dataset or file)|'collection'(dataset or container)|'dataset'|'container'|'file'
        :param long: Long format option to display more information for each DID.
        :param result: Recursively list DIDs content.
        """

        result = []
        result = self._rucio.ListDids(scope=scope,
                                      filters=filters,
                                      type=type,
                                      long=long,
                                      recursive=recursive)

        return result


    def GetRule(self, upload_structure=None, rse=None, level=-1):
        """This function checks if for a given upload structure or Rucio DID a requested
        upload destination rule exists in Rucio already and returns a standardized
        dictionary.

        :param: upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param: rse: A valid Rucio Storage Element (RSE)
        :param: level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                which the 'did' is chosen from.
        :return rule: A dictionary of pre-definied rule information.
        """

        #analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(upload_structure, level)

        r_rules = self._rucio.ListDidRules(val_scope, val_dname)
        if len(r_rules) == 0:
            return self._rule_status_dictionary()

        rule = self._rule_status_dictionary()
        for i_rule in r_rules:

            if rse!= None and i_rule['rse_expression'] != rse:
                continue
            rule['rse'] = i_rule['rse_expression']
            rule['exists'] = True
            rule['state']  = i_rule['state']
            rule['cnt_ok'] = i_rule['locks_ok_cnt']
            rule['cnt_repl'] = i_rule['locks_replicating_cnt']
            rule['cnt_stuck'] = i_rule['locks_stuck_cnt']
            rule['id'] = i_rule['id']
            #print("_", i_rule['expires_at'], type(i_rule['expires_at']))
            if i_rule['expires_at'] == None:
                rule['expires'] = None
            else:
                rule['expires'] = i_rule['expires_at'].strftime("%Y-%m-%d-%H:%M:%S")

        return rule

    def CheckRule(self, upload_structure=None, rse=None, level=-1):
        """Check the status message for a Rucio DID or dictionary template rule.
        This is a shortcut in combination with the memberfunction GetRule()

        :param: upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary. The member
                function GetRule(...) evaluates the upload_structure variable.
        :param: rse: Specify a valid Rucio Storage Element (RSE)
        :param: level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                which the 'did' is chosen from.
        :return: r_status: Rucio rule status: OK, STUCK or REPLICATING. If rule does not exists returns NoRule
        """

        if rse == None:
            return "NoRule"

        rule = self.GetRule(upload_structure, rse, level)
        #This function checks if for a given upload structure and a requested
        #upload destination already a rule exists in Rucio

        #HINT: GetRule returns None if no rule found now!
        #print(rule)

        r_status = None
        if rule['exists'] == False:
            r_status = "NoRule"
        elif rule['exists'] == True and rule['state'] == 'OK' and rule['cnt_ok'] > 0 and rule['cnt_repl'] == 0 and rule['cnt_stuck'] == 0:
            r_status = "OK"
        elif rule['exists'] == True and rule['state'] == 'REPLICATING':
            r_status = "REPLICATING"
        elif rule['exists'] == True and rule['state'] == 'STUCK':
            r_status = "STUCK"

        return r_status


    def VerifyLocations(self, upload_structure=None, upload_path=None, checksum_test=False, level=-1):
        """This function checks if for a given upload structure or Rucio DID a requested
        upload destination rule exists in Rucio already and returns a standardized
        dictionary.

        :param upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param upload_path: A path which holds the files for the Rucio upload
        :param checksum_test: Enable extended checksum test with True.
        :param level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                which the 'did' is chosen from.
        :return (check_success, diff_rucio, diff_disk): check_success True if the same files in Rucio such on disk
                                                        otherwise False.
                                                        diff_rucio returns a list of files which are in Rucio but not
                                                        on disk.
                                                        diff_disk returns a list of files which are on disk but not
                                                        attached to the Rucio DID.
        """

        #analyse the function input regarding its allowed definitions:
        val_scope, val_dname = self._VerifyStructure(upload_structure, level)

        #get all files from the physical location:
        list_folder_phys = []
        list_dirpath_phys = []
        list_files_phys = []
        for (dirpath, dirnames, filenames) in os.walk(upload_path):
            list_dirpath_phys.extend(dirpath)
            list_folder_phys.extend(dirnames)
            list_files_phys.extend(filenames)
            break

        nb_files_phys = len(list_files_phys)


        #list all attached files from the rucio cataloge:
        list_files_rucio = self._rucio.ListContent(val_scope, val_dname)

        rucio_files = []
        for i_file in list_files_rucio:
            rucio_files.append(i_file['name'])
        nb_files_rucio = len(rucio_files)

        #calculate diff of the filename lists
        #Files which are Rucio but not on disk:
        diff_rucio = list(set(rucio_files) - set(list_files_phys))
        #Files which are on disk but not in Rucio
        diff_disk  = list(set(list_files_phys) - set(rucio_files))

        #ToDo: Need a checksum test
        if checksum_test == True:
            print("Implement a checksum test")
            pass

        if nb_files_phys == nb_files_rucio:
            return (True, diff_rucio, diff_disk)
        else:
            return (False, diff_rucio, diff_disk)

    #Rucio download section
    def DownloadDids(self, dids=None, download_path=".", rse=None,
                     no_subdir=False, transfer_timeout=None,
                     num_threads=3, trace_custom_fields={}):
        """Function: DownloadDids(...)

        This functions offers to download a list if Rucio DIDs which are given by a list.

        :param dids: A string or a list of strings which follows the Rucio DID nameing convention
        :param download_path: Path to store the downloaded data
        :param rse: Specify the RSE from where the data are going to be downloaded
        :param no_subdir: True if no sub directory is going to be created.
        :param transfer_timeout: Wait for so many seconds and try to continue with the download (optional)
        :param num_threads: Standard two (2) threads are used for downloading on a CPU (optional)
        :param trace_custom_fields: Customize download, see Rucio tutorials optional)
        :return result: A list of Rucio download information as a list of dictionaries. If it fails 1
        """

        #Return failure if no dids information is given:
        if dids == None:
            return 1

        #To enforce the correct input on dids:
        if isinstance(dids, str):
            dids = [dids]

        #Create the list of dictionaries for download
        dw = []
        for i_did in dids:

            dw_dict = {}
            dw_dict['did'] = i_did
            dw_dict['rse'] = rse
            dw_dict['base_dir'] = download_path
            dw_dict['no_subdir'] = no_subdir
            dw_dict['transfer_timeout'] = transfer_timeout
            dw.append(dw_dict)

        #Download from Rucio
        result = self._rucio.DownloadDids(items=dw,
                                          num_threads=num_threads,
                                          trace_custom_fields=trace_custom_fields)

        return result

    def DownloadChunks(self, download_structure=None, chunks=None, download_path=".",
                       rse=None, no_subdir=False, transfer_timeout=None,
                       num_threads=2, trace_custom_fields={}, level=-1):
        """Function: DownloadChunks(...)

        This function offers to download specific chunks from Rucio a specific DiD.
        Warning: This function is heavily made for XENON internal structures. Please use with care.

        :param download_structure: A valid Rucio DID (string) or a template dictionary of an existing DID
        :param download_path: Path to store the downloaded data
        :param rse: Specify the RSE from where the data are going to be downloaded
        :param chunks: A list (strings) chunk numbers (Format: XXXXXX (six digits), or metadata.json:
                       Example: list=['000000', '000001', 'metadata.json']
        :param no_subdir: True if no sub directory is going to be created.
        :param transfer_timeout: Wait for so many seconds and try to continue with the download (optional)
        :param num_threads: Standard two (2) threads are used for downloading on a CPU (optional)
        :param trace_custom_fields: Customize download, see Rucio tutorials optional)
        :param level: Specify the download DID from a template dictionary if it is hand over (optional, the last level
                      is chosen)
        :return result: A list of Rucio download information as a list of dictionaries. If it fails 1
        """

        #Begin with analysing the download structure (similar to upload_structure)
        val_scope, val_dname = self._VerifyStructure(download_structure, level)

        dw = []
        for i_chunk in chunks:
            dw_dict = {}
            dw_dict['did'] = "{0}:{1}-{2}".format(val_scope, val_dname, i_chunk)
            dw_dict['rse'] = rse
            dw_dict['base_dir'] = download_path
            dw_dict['no_subdir'] = no_subdir
            dw_dict['transfer_timeout'] = transfer_timeout
            dw.append(dw_dict)

        result = self._rucio.DownloadDids(dw, num_threads, trace_custom_fields)

        return result

    def Download(self, download_structure=None, download_path=".",
                 rse=None, no_subdir=False, transfer_timeout=None,
                 num_threads=2, trace_custom_fields={}, level=-1):
        """Function: Download(...)

        This function offers to download from Rucio a specific DiD.

        :param download_structure: A valid Rucio DID (string) or a template dictionary of an existing DID
        :param download_path: Path to store the downloaded data
        :param rse: Specify the RSE from where the data are going to be downloaded
        :param no_subdir: True if no sub directory is going to be created.
        :param transfer_timeout: Wait for so many seconds and try to continue with the download (optinal)
        :param num_threads: Standard two (2) threads are used for downloading on a CPU (optinal)
        :param trace_custom_fields: Customize download, see Rucio tutorials (optinal)
        :param level: Specify the download DID from a template dictionary if it is hand over (optional, the last level
                      is chosen)
        :return result: A list of Rucio download information as a list of dictionaries. If it fails 1
        """

        #Begin with analysing the download structure (similar to upload_structure)
        val_scope, val_dname = self._VerifyStructure(download_structure, level)

        #Prepare a download dictionary:
        dw_dict = {}
        dw_dict['did'] = "{0}:{1}".format(val_scope, val_dname)
        dw_dict['rse'] = rse
        dw_dict['base_dir'] = download_path
        dw_dict['no_subdir'] = no_subdir
        dw_dict['transfer_timeout'] = transfer_timeout

        result = self._rucio.DownloadDids(dw_dict, num_threads, trace_custom_fields)

        return result


    #----Rucio upload section

    def UploadToScope(self, scope=None, upload_path=None,
                      rse=None, rse_lifetime=None):
        """Function: UploadToScope()

        Upload a folder to a Rucio scope

        :param scope: A string which follows the rules of Rucio string
        :param upload_path: A valid (string) to a folder which holds a file (or files) for upload
        :param rse: A valid Rucio Storage Element (RSE)
        :param rse_lifetime: A valid (int) which defines the lifetime of the transfer rule after upload.
        :return result: (upload_status, rse_rule) means:\n
                * (0, {'result': 0, 'lifetime': rse_lifetime}) for success and applied lifetime to the rule\n
                * (0, 1) for success and no rse_lifetime to the rule\n
                * (1, 1) for upload failure and rse_lifetime is not given\n
                * (1, {'result':1, 'lifetime': rse_lifetime}) for upload failure and rse_lifetime is skipped automatically
        """

        result = 1
        result_rule = 1

        if isinstance(scope, str) == False:
            print("Function UploadToScope() needs an Rucio (str) scope as input")
            exit(1)
        if rse==None:
            print("No Rucio Storage Element (rse) given.")
            exit(1)
        if upload_path ==None:
            print("No path/file given for upload")
            exit(1)

        #get all files from the physical location:
        list_files_phys = []
        for (dirpath, dirnames, filenames) in os.walk(upload_path):
            list_files_phys.extend(filenames)
            break

        #Create Scope:
        result = self._rucio.CreateScope(account=self.rucio_account,
                                scope=scope)
        #result 0 - success or 1 - failure

        upload_dict = {}
        upload_dict['path'] = upload_path + "/"
        upload_dict['rse'] = rse
        upload_dict['did_scope'] = scope

        result = self._rucio.Upload(upload_dict=[upload_dict])

        if rse_lifetime != None and isinstance(rse_lifetime, int) and \
            rse_lifetime > 0:# and \
            #result==0:
            # create an UpdateRule() conform input:
            rule = [f"rucio-catalogue:{rse}:{rse_lifetime}"]

            #apply continuously:
            result_rule_count = []
            for i_file in list_files_phys:
                upload_structure=f"{scope}:{i_file}"
                result_count = self.UpdateRules(upload_structure=upload_structure, rse_rules=rule)
                #evaluate result
                result_rule_count.append(result_count.get(rse,1))
            k_count = 0
            for ik in result_rule_count:
                k_count+=ik['result']

            if k_count == 0:
                result_rule = 0

        return (result, result_rule)

    def UploadToDid(self, upload_structure=None, upload_path=None,
               rse=None, rse_lifetime=None):
        """Function UploadToDid()

        This function uploads the content of given folder into a Rucio dataset
        which is identified by given DID.

        For example a folder:

        | /path/to/example/calibration_source_1
        |                   │
        |                   ├──18_t2_01
        |                   ├──18_t2_02
        |                   └──18_t2_03

        DID (dataset): calibration_data_day1:calibration_source_1

        Results a Rucio structure:

        | calibration_data_day1:calibration_source_1        (Rucio dataset)
        |           │
        |           ├──calibration_data_day1:18_t2_01   (Rucio file attached to dataset)
        |           ├──calibration_data_day1:18_t2_02   (Rucio file attached to dataset)
        |           └──calibration_data_day1:18_t2_03   (Rucio file attached to dataset)


        :param upload_structure: A Rucio DID (type str) in form of scope:name
        :param upload_path: A path to an existing folder with data for upload
        :param rse: A valid Rucio Storage Element (RSE) for the initial upload
        :param rse_lifetime: (Optional) A lifetime in seconds
        :return result: (upload_status, rse_rule) means:\n
                  * (0, {'result': 0, 'lifetime': rse_lifetime}) for success and applied lifetime to the rule\n
                  * (0, 1) for success and no rse_lifetime to the rule\n
                  * (1, 1) for upload failure and rse_lifetime is not give\n
                  * (1, {'result':1, 'lifetime': rse_lifetime}) for upload failure and rse_lifetime is skipped automatically
        """

        result = 1
        result_rule = 1

        #Go into
        if isinstance(upload_structure, str) == False and ":" not in upload_structure:
            print("Function UploadDid() needs an Rucio (str) DID as input (scope:name)")
            return (1, 1)
        if rse==None:
            print("No Rucio Storage Element (rse) given.")
            return (1, 1)
        if upload_path ==None:
            print("No path/file given for upload")
            return (1, 1)

        upload_scope = upload_structure.split(":")[0]
        upload_dname = upload_structure.split(":")[1]

        #Create Scope:
        result = self._rucio.CreateScope(account=self.rucio_account,
                                scope=upload_scope)
        #result 0 - success or 1 - failure

        #create the upload dictionary:
        upload_dict = {}
        upload_dict['path'] = upload_path + "/"
        upload_dict['rse'] = rse
        upload_dict['lifetime'] = rse_lifetime
        upload_dict['dataset_scope'] = upload_scope
        upload_dict['dataset_name'] = upload_dname
        upload_dict['did_scope'] = upload_scope

        result = self._rucio.Upload(upload_dict=[upload_dict])

        #We are ready here but we can apply a lifetime to the auto generated rule:
        #Obs! Rule lifetime only applied if upload was successful!
        if rse_lifetime != None and isinstance(rse_lifetime, int) and \
            rse_lifetime > 0 and \
            result==0:

            #create an UpdateRule() conform input:
            rule = [f"rucio-catalogue:{rse}:{rse_lifetime}"]
            #apply
            result_rule = self.UpdateRules(upload_structure=upload_structure, rse_rules=rule)
            #evaluate result
            result_rule = result_rule.get(rse, 1)

        return (result, result_rule)


    def Upload(self, did, upload_path, rse, lifetime=None):
        """Function: Upload(...)
        The data files of the upload_path are always uploaded to the last Rucio dataset.

        :param upload_structure: A string (Rucio DID form of "scope:name") or a template dictionary
        :param upload_path: The absolute path of your dataset
        :param rse: A valid Rucio Storage Element (RSE) for the upload
        :param rse_lifetime: The lifetime of the dataset (lowest level of template dictionary) after the upload
                             Hint: dataset lifetimes below 24h (86400 sec) are automatically set to 86400 sec.
        :param level: If a template dictionary is used, the level refers to the depth of the sorted dictionary at
                which the 'did' is chosen from.
        :return result: 0 for success, 1 for failure
        """


        scope, dataset = did.split(':')


        # create scope. If scope exists already, exception will be handled silently unless verbose=True is passed
        logger.debug(f"Creating scope {scope}, if it doesn't exist")
        self._rucio.CreateScope(account=self.rucio_account, scope=scope)

        # create dataset
        #logger.debug(f"Creating dataset {scope}:{dataset}")
        #self._rucio.CreateDataset(scope, dataset)

        # add a rule for this dataset
        #logger.debug(f"Creating a rule for {did} at {rse}")
        #self._rucio.AddRule([dict(scope=scope, name=dataset)], 1, rse, lifetime=lifetime)

        # smallest lifetime is 1 day?
        if lifetime != None and int(lifetime) < 86400:
            lifetime = 86400

        #Prepare an upload dictionary which is used later to upload to Rucio
        upload_dict = dict(path=upload_path,
                           rse=rse,
                           lifetime=lifetime,
                           did_scope=scope,
                           dataset_scope=scope,
                           dataset_name=dataset,
                           #register_after_upload=True
                           )

        # finally, upload the dataset
        result = self._rucio.Upload(upload_dict=[upload_dict])
        return result



import argparse
from admix.daemons.sync import SyncDaemon
from datetime import datetime, timedelta
from admix import __version__


def version(args):
    print(__version__)


def sync(args):
    query = SyncDaemon.query
    if args.run:
        query['number'] = args.run
    if args.days_ago:
        query['start'] = {'$gt': datetime.utcnow() - timedelta(args.days_ago)}

    syncer = SyncDaemon(db_query=query, dtype=args.dtype)
    syncer.single_loop(max_iterations=args.limit, progress_bar=args.progress)


def main():
    parser = argparse.ArgumentParser(description="Main admix command. Use this to call various subcommands.")

    subparsers = parser.add_subparsers(help='sub-command help')

    # version subcommand
    version_parser = subparsers.add_parser('version', help='Print version and exit')
    version_parser.set_defaults(func=version)

    # sync subcommand
    sync_parser = subparsers.add_parser('sync', help='admix sync: to sync rucio with the runsDB')
    sync_parser.add_argument('--run', type=int, help='Run number')
    sync_parser.add_argument('--days_ago', type=int, help='Sync runs taken in the last DAYS_AGO days')
    sync_parser.add_argument('--dtype', help='Only sync the DTYPE strax datatype')
    sync_parser.add_argument('--limit', help='Only sync a max of LIMIT runs', default=0)
    sync_parser.add_argument('--progress', action='store_true', help='Display progress')
    sync_parser.set_defaults(func=sync)

    # now do the thing
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



import sys
from tqdm import tqdm
from . import rucio
from .utils import make_did


def get_dtype_status(run, dtype, hash, rse=None):
    did = make_did(run, dtype, hash)
    if rse:
        rules = rucio.list_rules(did, rse_expression=rse)
    else:
        rules = rucio.list_rules(did)

    title = f"| {'RSE':^20s} | {'STATUS':^15s} | {'PROGRESS':^25s} |"
    suptitle = f"| {did:^66s} |"
    overline_bar = tqdm(desc="-" * len(title), bar_format="{desc}", file=sys.stdout)
    overline_bar.close()
    suptitle_bar = tqdm(desc=suptitle, bar_format="{desc}", file=sys.stdout)
    suptitle_bar.close()
    underline_bar = tqdm(desc="-" * len(title), bar_format="{desc}", file=sys.stdout)
    underline_bar.close()
    title_bar = tqdm(desc=title, bar_format="{desc}", file=sys.stdout)
    title_bar.close()
    underline_bar = tqdm(desc="-"*len(title), bar_format="{desc}", file=sys.stdout)
    underline_bar.close()
    for i, rule in enumerate(rules):
        num_ok = rule['locks_ok_cnt']
        num_replicating = rule['locks_replicating_cnt']
        num_stuck = rule['locks_stuck_cnt']
        total = num_ok + num_replicating + num_stuck

        rse = rule['rse_expression']
        bar = tqdm(total=total,
                   desc=f"| {rse:^20s} | {rule['state']:^15s} |",
                   bar_format="{desc} {percentage:3.0f}%|{bar}|{n_fmt}/{total_fmt}|",
                   ncols=70,
                   file=sys.stdout
                   )
        bar.update(num_ok)
        bar.close()
    underline_bar = tqdm(desc="-" * len(title), bar_format="{desc}", file=sys.stdout)
    underline_bar.close()
    # blank space for readability
    print()


def get_run_status(run, dtype=None, rse=None):
    scope = f"xnt_{run:06d}"
    datasets = rucio.list_datasets(scope)
    for d in datasets:
        dt, hsh = d.split('-')
        if dtype:
            if dt == dtype:
                get_dtype_status(run, dt, hsh, rse=rse)
        else:
            get_dtype_status(run, dt, hsh, rse=rse)



"""
.. module:: rucio_api
   :platform: Unix
   :synopsis: An extensive wrapper for the Rucio API in aDMIX

.. moduleauthor:: Boris Bauermeister <Boris.Bauermeister@gmail.com>

"""
#from __future__ import with_statement
import os

from admix import logger
import os

from rucio.client.client import Client
from rucio.client.uploadclient import UploadClient
# from admix.interfaces.uploadclient import UploadClient
from rucio.client.downloadclient import DownloadClient
from rucio.common.exception import DataIdentifierAlreadyExists
from rucio.common.exception import AccessDenied
from rucio.common.exception import Duplicate
from rucio.common.exception import DuplicateContent
from rucio.common.exception import DuplicateRule


class RucioAPI():
    """Class RucioAPI()

    This class presents an approach to collect all necessary Rucio calls
    in one class. That allows easy handling of Rucio calls based on the
    Rucio API.
    If there are ever changes in the Rucio API, here is the wrapper to fix it.
    """

    def __init__(self, enable_print=False):
        """Function: __init__()

        Constructor of RucioAPI class. Comes with a setting set up the print statements to terminal

        :param enable_print: If True then enable print to terminal
        """
        self._print_to_screen = enable_print
        self._rucio_ping = None
        self._rucio_account = os.environ.get("RUCIO_ACCOUNT")
        self.ConfigHost()

    def __del__(self):
        """Function: __del__()

        Destructor - No further description
        """
        pass

    # Here comes the backend configuration part:
    def SetRucioAccount(self, rucio_account=None):
        """Function: SetRucioAccount
        :param rucio_account: The Rucio account you would like to work with
        """

        self._rucio_account = rucio_account

    def SetConfigPath(self, config_path=None):
        """Function: SetConfigPath
        This option is only important for legacy command line support and
        is ignored in RucioAPI setup.
        :param config_path: Path to CLI configuration file
        """
        pass
    def SetProxyTicket(self, proxy_path=None):
        """Function: SetProxyTicket
        This option is only important for legacy command line support and
        is ignored in RucioAPI setup.
        :param proxy_path: Path to CLI configuration file
        """
        pass
    def SetHost(self, hostname=None):
        """Function: SetHost
        This option is only important for legacy command line support and
        is ignored in RucioAPI setup.
        :param hostname: Path to CLI configuration file
        """
        pass
    def ConfigHost(self):
        """Function: ConfigHost

        This member function setup the rucioAPI backend.
        To ensure full functionality, it needs:
        * Client()
        * UploadClient()
        * DownloadClient()

        :raise Exception if Rucio API is not ready (miss-configured)
        """
        try:
            self._rucio_client = Client()
            self._rucio_client_upload = UploadClient(logger=logger)
#            self._rucio_client_upload = UploadClient(tracing=False)
#            print("Tracing set to False")
            self._rucio_client_download = DownloadClient()
            self._rucio_ping = self._rucio_client.ping

        except:
            print("Can not init the Rucio API")
            print("-> Check for your Rucio installation")
            exit(1)


    # finished the backend configuration for the Rucio API

    def Whoami(self):
        """RucioAPI:Whoami
        Results a dictionary to identify the current
        Rucio user and credentials.
        """
        return self._rucio_client.whoami()

    def GetRucioPing(self):
        """Function: GetRucioPing
        :return If ConfigHost is executed without execption GetRucioPing provides a Rucio ping
        """

        return self._rucio_client.ping

    #The scope section:
    def CreateScope(self, account, scope, verbose=False):
        """Function: CreateScope()

        Create a new Rucio scope what does not yet exists yet.
        Be aware that you need Rucio permissions to do it. Check your Rucio account and settings.

        :param account: The Rucio account you are working with (need to be allowed to create scopes)
        :param scope: The scope name you like to create
        :return result:
        """

        result = 1
        try:
            self._rucio_client.add_scope(account, scope)
            result = 0
        except AccessDenied as e:
            print(e)
        except Duplicate as e:
            if verbose:
                print(e)
            else:
                pass
        return result

        #Several list commands

    def GetRSE(self, rse):
        """Function: GetRSE(...)

        Return further information about the RSE setup of a specific RSE
        :param rse: A (string) valid Rucio Storage Element (RSE) name
        :return result: A dictionary which holds information according the selected RSE
        """

        result = {}
        try:
            result = self._rucio_client.get_rse(rse)
        except:
            print("No RSE attributes received for {0}".format(rse))
        return result

    def ListRSEAttributes(self, rse):
        """Function: ListRSEAttributes(...)

        Return some attributes of a Rucio Storage Element
        Received keys are fts, fts_testing, RSE-NAME, istape

        :param rse: A valid (string) Rucio Storage Element (RSE) name
        :return result: A dictionary with RSE attributes
        """

        result = {}

        try:
            result = self._rucio_client.list_rse_attributes(rse)
        except:
            print("No RSE attributes received for {0}".format(rse))
        return result


    def ListRSEs(self):
        """Function: ListRSEs

        Returns an overview about all registered Rucio Storage elements in the current setup

        :return result: A list of dictionaries. Each dictionary holds RSE information. If not successful []
        """
        result = []
        try:
            result = list(self._rucio_client.list_rses())
        except:
            print("No RSE received from Rucio.")

        return result

    def ListContent(self, scope, name):
        """Function: ListContent()

        :param scope: A string which refers to a Rucio scope
        :param name: A string which refers to a Rucio name
        :return result: A list of dictionaries with the attached files to the DID
        """
        result = []
        try:
            return list(self._rucio_client.list_content(scope, name))
        except TypeError as e:
            print(e)

        return result

    def ListScopes(self):
        """Function: ListScopes()

        List all created scopes in the Rucio catalogue

        :return result: A list of scopes, otherwise []
        """
        result = []
        try:
            result = self._rucio_client.list_scopes()
        except:
            print("No scopes? - Check that!")
        return result

    def ListFileReplicas(self, scope, lfn):
        """Function: ListFileReplicas(...)

        List all your files which are attached to a dataset or container

        :param scope: A string which follows the rules of a Rucio scope
        :param lfn: the lfn.
        :return result: A list of file replicas, otherwise []
        """

        #todo FIX ME
        result = []
        result = self._rucio_client.list_file_replicas(scope, lfn)
        return result

        #try:
        #    result = self._rucio_client.list_file_replicas(scope, lfn)
        #except AttributeError as e:
        #    print(e)
        return result

    def ListFiles(self, scope, name, long=True):
        """Function: ListFiles(...)

        List all your files which are attached to a dataset or container

        :param scope: A string which follows the rules of a Rucio scope
        :param name: A string which follows the rules of a Rucio name
        :param long: Define another output (Check the Rucio tutorials for it)
        :return result: A list of files, otherwise []
        """
        result = []
        try:
            result = self._rucio_client.list_files(scope, name, long=None)
        except:
            print("No files are listed for {0}:{1}".format(scope, name))
        return result

    def ListDids(self,scope, filters, type='collection', long=False, recursive=False):
        """
        List all data identifiers in a scope which match a given pattern. Check Rucio github page for details

        :param scope: The valid string which follows the Rucio scope name.
        :param filters: A dictionary of key/value pairs like {'name': 'file_name','rse-expression': 'tier0'}.
        :param type: The type of the did: 'all'(container, dataset or file)|'collection'(dataset or container)|'dataset'|'container'|'file'
        :param long: Long format option to display more information for each DID.
        :param result: Recursively list DIDs content.
        """

        result = []
        try:
            return list(self._rucio_client.list_dids(scope, filters, type, long, recursive))
        except TypeError as e:
            print(e)
        return result


    def ListDidRules(self, scope, name):
        """Return a class generator from Rucio which contains the
        individual rules to iterate over (or to create a list from)

        :param scope: A string which refers to the Rucio scope
        :param name: A string which refers to the Rucio name (a container, dataset or file name)

        :return: A list of Rucio transfer rules with additional rule information. Each list element stands for a
                 Rucio Storage Element (RSE). List is empty if not successful or nor rules.
        """

        result = []
        try:
            return list(self._rucio_client.list_did_rules(scope, name))
        except TypeError as e:
            print(e)

        return result

    #Attach and detach:
    def AttachDids(self, scope, name, attachment, rse=None):
        """Function: AttachDids(...)

        This function allows to attach datasets or containers to a top-level dataset or container.
        The parameters scope and name define the top-level structure (container or dataset) and the dictionary or
        the list of dictionaries contains the information about what is attached to the top-level structure.

        More information under https://github.com/rucio/rucio


        :param scope: A string which follows the rules of a Rucio scope
        :param name: A string which follows the rules of a Rucio name
        :param attachment: A dictionary or a list of dictionaries which consist of two keys: scope and name
                           example{'scope': 'example_scope1', 'name':'example_name1'}
        :param rse: The RSE name when registering replicas. (optional)
        :return result: 0 if successful, 1 for failure
        """
        result = 1

        #In case there is only an individual dictionary provided, the dictionary is transformed into a list of
        #dictionaries.
        if isinstance(attachment, dict) == True:
            attachment = [attachment]

        #self._rucio_client.attach_dids(scope, name, attachment, rse=rse)

        try:
            self._rucio_client.attach_dids(scope, name, attachment, rse=rse)
            result = 0
        except DuplicateContent as e:
            print(e)

        return result

    def DetachDids(self, scope, name, dids):
        try:
            self._rucio_client.detach_dids(scope, name, dids)
        except:
            return None

    #Container and Dataset managment:
    def CreateContainer(self, scope, name, statuses=None, meta=None, rules=None, lifetime=None):
        """Function CreateContainer(...)

        Follows the Rucio API to create a Rucio container based on scope and container name. It accept also further
        Rucio features.
        More information under https://github.com/rucio/rucio

        :param scope: A string which follows the rules of a Rucio scope
        :param name: A string which follows the rules of a Rucio container name
        :param statuses: Status (optional)
        :param meta: Put in further meta data which are going to be connected to the container. (optional)
        :param rules: Define transfer rules which apply to the container immediately. (optional)
        :param lifetime: Set a Rucio lifetime to the container if you with (optional)
        :return result: 0 if successful, 1 for failure
        """
        result = 1
        try:
            self._rucio_client.add_container(scope, name, statuses=None, meta=None, rules=None, lifetime=None)
            result = 0
        except DataIdentifierAlreadyExists as e:
            print(e)
        return result

    def CreateDataset(self, scope, name, statuses=None, meta=None,
                      rules=None, lifetime=None, files=None, rse=None, verbose=False):
        """Function CreateDataset(...)

        Follows the Rucio API to create a Rucio dataset based on scope and dataset name. It accept also further
        Rucio features.
        More information under https://github.com/rucio/rucio

        :param scope:    A string which follows the rules of a Rucio scope
        :param name:     A string which follows the rules of a Rucio dataset name
        :param statuses: Status (optional)
        :param meta:     Put in further meta data which are going to be connected to the container. (optional)
        :param rules:    Define transfer rules which apply to the container immediately. (optional)
        :param lifetime: Set a Rucio lifetime to the container if you with (optional)
        :param verbose:  Flag to print DataIdentifierAlreadyExists exceptions
        :return result:  0 if successful, 1 for failure
        """
        result = 1
        try:
            self._rucio_client.add_dataset(scope, name, statuses=None, meta=None, rules=None, lifetime=None,\
                                           files=None, rse=None)
            result = 0
        except DataIdentifierAlreadyExists as e:
            if verbose:
                print(e)
        return result

    #Rules:
    def AddRule(self, dids, copies, rse_expression, weight=None, lifetime=None, grouping='DATASET', account=None,
                locked=False, source_replica_expression=None, activity=None, notify='N', purge_replicas=False,
                ignore_availability=False, comment=None, ask_approval=False, asynchronous=False, priority=3, meta=None):
        """Function: AddRule(...)

        A function to add a Rucio transfer rule to the given Rucio data identifiers (DIDs)
        More information under https://github.com/rucio/rucio

        :param dids:                       The data identifier set.
        :param copies:                     The number of replicas.
        :param rse_expression:             Boolean string expression to give the list of RSEs.
        :param weight:                     If the weighting option of the replication rule is used, the choice of RSEs takes their weight into account.
        :param lifetime:                   The lifetime of the replication rules (in seconds).
        :param grouping:                   ALL -  All files will be replicated to the same RSE.
                                           DATASET - All files in the same dataset will be replicated to the same RSE.
                                           NONE - Files will be completely spread over all allowed RSEs without any grouping considerations at all.
        :param account:                    The account owning the rule.
        :param locked:                     If the rule is locked, it cannot be deleted.
        :param source_replica_expression:  RSE Expression for RSEs to be considered for source replicas.
        :param activity:                   Transfer Activity to be passed to FTS.
        :param notify:                     Notification setting for the rule (Y, N, C).
        :param purge_replicas:             When the rule gets deleted purge the associated replicas immediately.
        :param ignore_availability:        Option to ignore the availability of RSEs.
        :param ask_approval:               Ask for approval of this replication rule.
        :param asynchronous:               Create rule asynchronously by judge-injector.
        :param priority:                   Priority of the transfers.
        :param comment:                    Comment about the rule.
        :param meta:                       Metadata, as dictionary.

        :return result:  0 if successful, 1 for failure
        """
        result = 1

        try:
#            self._rucio_client.add_replication_rule(dids, copies, rse_expression, weight=None, lifetime=lifetime,
#                                                    grouping='DATASET', account=None, locked=False,
#                                                    source_replica_expression=None, activity=None, notify='N',
#                                                    purge_replicas=False, ignore_availability=False, comment=None,
#                                                    ask_approval=False, asynchronous=False, priority=3)
            self._rucio_client.add_replication_rule(dids, copies, rse_expression, weight=None, lifetime=lifetime,
                                                    grouping='DATASET', account=None, locked=False,
                                                    source_replica_expression=source_replica_expression, activity=None, notify='N',
                                                    purge_replicas=False, ignore_availability=False, comment=None,
                                                    ask_approval=False, asynchronous=False, priority=priority)
            result = 0
        except DuplicateRule as e:
            print(e)

        return result

    def UpdateRule(self, rule_id, options=None):
        """Function UpdateRule()

        Aims to update a particular rule according to its rule_id and further option such as lifetime
        :param rule_id: A Rucio rule id string
        :param options: A dictionary with certain options (e.g. lifetime, weight, ,...)
        :return result: 0 on success, 1 at failure
        """
        result = 1
        try:
            self._rucio_client.update_replication_rule(rule_id, options)
            result = 0
        except:
            print("Raised exception in UpdateRule")

        return result

    def GetReplicationRule(self, rule_id, estimate_ttc=False):
        """Function: GetReplicationRule(...)

        Get information on the replication rule based on the rule ID

        :param rule_id: A valid Rucio rule ID
        :return result: Information on the replication rule, otherwise 1
        """
        result = 1
        try:
            result = self._rucio_client.get_replication_rule(self, rule_id, estimate_ttc=False)
        except:
            print("No replication rule to get")
        return result

    def DeleteRule(self, rule_id):
        """Function: DeleteRule(...)

        Deletes a replication rule.
        :param rule_id: A rucio rule id string
        """
        self._rucio_client.delete_replication_rule(rule_id, purge_replicas=True)

    #Metadata:
    def GetMetadata(self, scope, name):
        try:
            return self._rucio_client.get_metadata(scope, name)
        except:
            return None

    def SetMetadata(self, scope, name, key, value, recursive=False):
        try:
            return self._rucio_client.set_metadata(scope, name, key, value, recursive=False)
        except:
            return None

    #Data upload / download / register
    def Upload(self, upload_dict=None):
        """Function: Upload()

        The list of dictionaries need to follow this convention:
        Rucio/Github: https://github.com/rucio/rucio/blob/master/lib/rucio/client/uploadclient.py#L71

        :param upload_dict: A list object with dictionaries
        :return result: 0 on success, 1 on failure

        """
        result = self._rucio_client_upload.upload(upload_dict)
        return result

    def DownloadDids(self, items, num_threads=2, trace_custom_fields={}):
        """Function: DownloadDids(...)

        Download from the Rucio catalogue by Rucio DIDs (or a list of them)

        :param items: A list or a dictionary of information what to download
        :param num_threads: Specify the number threads on the CPU, standard 2 (optional)
        :param trace_custom_fields: Customize downloads (Look at Rucio tutorials) (optional)
        :return result: A list of dictionaries of Rucio download result messages. If it fails: 1
        """
        result = 1
        #if a dictionary is handed over, we create a list of it.
        if isinstance(items, dict):
            items = [items]

        try:
            result = self._rucio_client_download.download_dids(items=items,
                                                       num_threads=num_threads,
                                                       trace_custom_fields=trace_custom_fields)
        except:
            result = 1

        return result

    def Register(self, rse, files, ignore_availability=True):
        #See email "IceCube Script to register data"
        #from Benedikt.
        #files = {
                #'scope': self.scope,
                #'name': replicas[filemd]['name'],
                #'adler32': replicas[filemd]['adler32'],
                #'bytes': replicas[filemd]['size'],
               #} for filemd in replicas]
               #--> Think about metadata
        try:
            self._rucio_client.add_replicas(rse, files, ignore_availability)
        except:
            print("Problem with file name does not match pattern")


        for filemd in replicas:
            try:
                self.didc.attach_dids(scope=self.scope, name=self.run_Number, dids=[{
                    'scope': self.scope,
                    'name': replicas[filemd]['name']}])
            except FileAlreadyExists:
                print("File already attached")
